import pytest


def test_decorator(clean_registry_by_name, clean_registry_by_model):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo")
    class Foo:
        pass

    expected = {
        "can_create": True,
        "can_list": True,
        "can_message": True,
        "can_read": True,
        "can_update": True,
        "model": Foo,
        "name": "Foo",
        "plural": "foos",
        "resource_id": "id",
        "url_part": "foo",
    }
    assert clean_registry_by_name["Foo"] == expected
    assert clean_registry_by_model[Foo] == expected

    with pytest.raises(ValueError):

        @register_model("Foo", url_part="foo")
        class AnotherFoo:
            pass

    @register_model(
        "Bar",
        url_part="baar",
        plural="barses",
        can_create=False,
        can_read=False,
        can_update=False,
        can_list=False,
        can_message=False,
    )
    class Bar:
        pass

    expected = {
        "can_create": False,
        "can_list": False,
        "can_message": False,
        "can_read": False,
        "can_update": False,
        "model": Bar,
        "name": "Bar",
        "plural": "barses",
        "resource_id": "id",
        "url_part": "baar",
    }
    assert clean_registry_by_name["Bar"] == expected
    assert clean_registry_by_model[Bar] == expected


def test_lookup_model_by_name(clean_registry_by_name, clean_registry_by_model):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo")
    class Foo:
        pass

    from edutap.wallet_google.registry import lookup_model_by_name

    assert lookup_model_by_name("Foo") == Foo

    with pytest.raises(KeyError):
        lookup_model_by_name("Bar")


def test_lookup_metadata_by_name(clean_registry_by_name, clean_registry_by_model):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo")
    class Foo:
        pass

    from edutap.wallet_google.registry import lookup_metadata_by_name

    assert lookup_metadata_by_name("Foo") == {
        "plural": "foos",
        "resource_id": "id",
        "can_create": True,
        "can_list": True,
        "can_message": True,
        "can_read": True,
        "can_update": True,
        "model": Foo,
        "name": "Foo",
        "url_part": "foo",
    }

    with pytest.raises(KeyError):
        lookup_metadata_by_name("Bar")


def test_lookup_metadata_by_model_instance(
    clean_registry_by_name, clean_registry_by_model
):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo")
    class Foo:
        pass

    foo = Foo()

    from edutap.wallet_google.registry import lookup_metadata_by_model_instance

    assert lookup_metadata_by_model_instance(foo) == {
        "plural": "foos",
        "resource_id": "id",
        "can_create": True,
        "can_list": True,
        "can_message": True,
        "can_read": True,
        "can_update": True,
        "model": Foo,
        "name": "Foo",
        "url_part": "foo",
    }

    class Bar:
        pass

    bar = Bar()

    with pytest.raises(KeyError):
        lookup_metadata_by_model_instance(bar)


def test_raise_when_operation_not_allowed(
    clean_registry_by_name, clean_registry_by_model
):
    from edutap.wallet_google.registry import register_model

    @register_model("foo", url_part="foo", can_message=False)
    class Foo:
        pass

    from edutap.wallet_google.registry import raise_when_operation_not_allowed

    assert raise_when_operation_not_allowed("foo", "create") is None

    with pytest.raises(ValueError):
        raise_when_operation_not_allowed("foo", "message")


def test_lookup_model_by_plural_name(clean_registry_by_name):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo", plural="fooos")
    class Foo:
        pass

    from edutap.wallet_google.registry import lookup_model_by_plural_name

    assert lookup_model_by_plural_name("fooos") == Foo

    with pytest.raises(LookupError):
        lookup_model_by_plural_name("barses")


def test_lookup_metadata_by_model_type(clean_registry_by_model):
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo", plural="foos")
    class Foo:
        pass

    from edutap.wallet_google.registry import lookup_metadata_by_model_type

    md = lookup_metadata_by_model_type(Foo)
    assert md["name"] == "Foo"
    assert md["url_part"] == "foo"
    assert md["plural"] == "foos"



from ..bases import Model


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Review(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/Review
    """

    comments: str | None = None



"""
Models to be used to assemble the JWT for the save link (add to wallet link).
"""

from ..bases import Model
from ..passes import generic
from ..passes import retail
from ..passes import tickets_and_transit
from ..passes.bases import Reference
from datetime import datetime


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class JWTPayload(Model):

    eventTicketClasses: (
        list[tickets_and_transit.EventTicketClass | Reference] | None
    ) = None
    eventTicketObjects: (
        list[tickets_and_transit.EventTicketObject | Reference] | None
    ) = None
    flightClasses: list[tickets_and_transit.FlightClass | Reference] | None = None
    flightObjects: list[tickets_and_transit.FlightObject | Reference] | None = None
    giftCardClasses: list[retail.GiftCardClass | Reference] | None = None
    giftCardObjects: list[retail.GiftCardObject | Reference] | None = None
    loyaltyClasses: list[retail.LoyaltyClass | Reference] | None = None
    loyaltyObjects: list[retail.LoyaltyObject | Reference] | None = None
    offerClasses: list[retail.OfferClass | Reference] | None = None
    offerObjects: list[retail.OfferObject | Reference] | None = None
    transitClasses: list[tickets_and_transit.TransitClass | Reference] | None = None
    transitObjects: list[tickets_and_transit.TransitObject | Reference] | None = None
    genericClasses: list[generic.GenericClass | Reference] | None = None
    genericObjects: list[generic.GenericObject | Reference] | None = None


class JWTClaims(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/Jwt

    Note: `exp` is added to the model as standard field of the JWT specification,
    see https://datatracker.ietf.org/doc/html/rfc7519#section-4.1.4,
    but it is not documented in the Google Wallet JWT Specification!
    """

    iss: str
    aud: str = "google"
    typ: str = "savetowallet"
    iat: str | datetime = ""
    payload: JWTPayload
    origins: list[str]

    # a standard field in the JWT Specification, but not documented for the Google Wallet JWT:
    exp: str | datetime = ""



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from .enums import AnimationType
from .enums import NfcConstraint
from .enums import ScreenshotEligibility
from .localized_string import LocalizedString
from pydantic import AnyHttpUrl
from pydantic import AnyUrl
from pydantic import Field
from typing import Annotated
from typing_extensions import deprecated


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Uri(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Uri
    """

    # inherits kind (deprecated)
    uri: AnyUrl | str | None = None
    description: str | None = None
    localizedDescription: LocalizedString | None = None
    id: str | None = None


class ImageUri(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Image#imageuri
    """

    uri: AnyUrl
    description: Annotated[
        str | None,
        Field(
            deprecated=deprecated(
                'The Attribute "description" is deprecated on "ImageUri", use "contentDesceription" on parent "Image" Instance.'
            ),
            default=None,
            exclude=True,
        ),
    ]
    localizedDescription: Annotated[
        LocalizedString | None,
        Field(
            deprecated=deprecated(
                'The Attribute "localizedDescription" is deprecated on "ImageUri", use "contentDesceription" on parent "Image" Instance.'
            ),
            default=None,
            exclude=True,
        ),
    ]


class Image(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Image
    """

    # inherits kind (deprecated)
    sourceUri: ImageUri
    contentDescription: LocalizedString | None = None


class PassConstraints(Model):
    """
    see https://developers.google.com/wallet/generic/rest/v1/PassConstraints
    """

    screenshotEligibility: ScreenshotEligibility = (
        ScreenshotEligibility.SCREENSHOT_ELIGIBILITY_UNSPECIFIED
    )
    nfcConstraint: list[NfcConstraint] | None = Field(
        default_factory=lambda: [
            NfcConstraint.NFC_CONSTRAINT_UNSPECIFIED,
        ]
    )


class SecurityAnimation(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/SecurityAnimation
    """

    animationType: AnimationType = AnimationType.ANIMATION_UNSPECIFIED


class GroupingInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/GroupingInfo
    """

    sortIndex: int | None = None
    groupingId: str | None = None


class Pagination(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Pagination
    """

    # inherits kind (deprecated)
    resultsPerPage: int
    nextPageToken: str | None = None


class CallbackOptions(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/CallbackOptions
    """

    url: AnyHttpUrl | None = None
    updateRequestUrl: Annotated[
        AnyHttpUrl | None,
        Field(
            deprecated=deprecated(
                'The Parameter "updateRequestUrl" is deprecated on "CallbackOption", use "url" instead.'
            ),
            default=None,
            exclude=True,
        ),
    ]


class SaveRestrictions(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/SaveRestrictions
    """

    restrictToEmailSha256: str | None = None



from edutap.wallet_google.api import _validate_data_and_convert_to_json
from edutap.wallet_google.models.passes.bases import ClassModel


def test_api__validate_data_and_convert_to_json__create_mode():
    model = ClassModel(id="12345", enableSmartTap=None)
    _, json = _validate_data_and_convert_to_json(ClassModel, model)
    assert json == '{"id":"12345"}'


def test_api__validate_data_and_convert_to_json__existing_mode():
    model = ClassModel(id="12345", enableSmartTap=None)
    _, json = _validate_data_and_convert_to_json(ClassModel, model, existing=True)
    assert json == '{"id":"12345","enableSmartTap":null}'



"""
Parts of this module are rewrites and borrows from from https://github.com/yoyowallet/google-pay-token-decryption
The above packages does not fulfill the needs we have here, but was a great starting point.
Copyright is by its original authors at Yoyo Wallet <dev@yoyowallet.com>
This file is under the MIT License, as found here https://github.com/yoyowallet/google-pay-token-decryption/blob/5cd006da9687171c1e35b55507b671c6e4eb513d/pyproject.toml#L8
"""

from ..models.handlers import CallbackData
from ..models.handlers import IntermediateSigningKey
from ..models.handlers import RootSigningPublicKeys
from ..models.handlers import SignedKey
from ..models.handlers import SignedMessage
from ..session import session_manager
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePrivateKey
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
from cryptography.hazmat.primitives.serialization import load_der_private_key
from cryptography.hazmat.primitives.serialization import load_der_public_key
from typing import cast

import base64
import httpx
import time


PROTOCOL_VERSION = "ECv2SigningOnly"
ALGORITHM = ECDSA(hashes.SHA256())
GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_URL = {
    # see https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#root-signing-keys
    "testing": "https://payments.developers.google.com/paymentmethodtoken/test/keys.json",
    "production": "https://payments.developers.google.com/paymentmethodtoken/keys.json",
}
GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE: dict[str, RootSigningPublicKeys] = {}


def google_root_signing_public_keys(google_environment: str) -> RootSigningPublicKeys:
    """
    Fetch Googles root signing keys once for the configured environment and return them or the cached value.
    """
    if GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE.get(google_environment, None) is not None:
        return GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE[google_environment]
    # fetch once
    resp = httpx.get(GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_URL[google_environment])
    resp.raise_for_status()
    GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE[google_environment] = (
        RootSigningPublicKeys.model_validate_json(resp.text)
    )
    return GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE[google_environment]


def _raw_private_key(in_key: str) -> str:
    """
    Returns the raw private key.
    """
    result = ""
    for line in in_key.splitlines():
        if "BEGIN PRIVATE KEY" in line:
            continue
        if "END PRIVATE KEY" in line:
            break
        result += line.strip()
    return result


def _construct_signed_data(*args: str) -> bytes:
    """
    Construct the signed message from the list of its components by concatenating the
    byte length of each component in 4 bytes little-endian format plus the UTF-8 encoded
    component.

    See https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#verify-signature
    or  https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#how-to-construct-the-byte-string-for-intermediate-signing-key-signature
    for an example.
    """
    signed = b""
    for a in args:
        signed += len(a).to_bytes(4, byteorder="little")
        signed += bytes(a, "utf-8")
    return signed


def _load_public_key(key: str) -> EllipticCurvePublicKey:
    derdata = base64.b64decode(key)
    return cast(
        EllipticCurvePublicKey,
        load_der_public_key(derdata, default_backend()),
    )


def _load_private_key(key: str) -> EllipticCurvePrivateKey:
    derdata = base64.b64decode(key)
    return cast(
        EllipticCurvePrivateKey,
        load_der_private_key(derdata, None, default_backend()),
    )


def _verify_intermediate_signing_key(
    public_keys: RootSigningPublicKeys,
    intermediate_signing_key: IntermediateSigningKey,
) -> bool:
    """Check the intermediate signing keys signature against the Google root public keys.

    see https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#verify-signature
    """
    signatures = [
        base64.decodebytes(bytes(sig, "utf-8"))
        for sig in intermediate_signing_key.signatures
    ]
    signed_data = _construct_signed_data(
        "Google",
        PROTOCOL_VERSION,
        intermediate_signing_key.signedKey,
    )
    for pkey in public_keys.keys:

        if pkey.protocolVersion != PROTOCOL_VERSION:
            continue

        public_key = _load_public_key(pkey.keyValue)
        for signature in signatures:
            try:
                public_key.verify(signature, signed_data, ALGORITHM)
            except (ValueError, InvalidSignature):
                # Invalid signature. Try the other signatures.
                ...
            else:
                # Valid signature was found
                return True
    return False


def verified_signed_message(data: CallbackData) -> SignedMessage:
    """
    Verifies the signature of the callback data.
    and returns the parsed SignedMessage
    """
    # parse message
    message = SignedMessage.model_validate_json(data.signedMessage)

    # get issuer_id
    if not message.classId or "." not in message.classId:
        raise ValueError("Missing classId")
    issuer_id = message.classId.split(".")[0]

    # shortcut if signature validation is disabled
    settings = session_manager.settings
    if settings.handler_callback_verify_signature == "0":
        return message

    if data.protocolVersion != PROTOCOL_VERSION:
        raise ValueError("Invalid protocolVersion")

    # check intermediate signing keys signature
    if not _verify_intermediate_signing_key(
        google_root_signing_public_keys(settings.google_environment),
        data.intermediateSigningKey,
    ):
        raise ValueError("Invalid intermediate signing key")

    # check intermediate signing keys expriration date
    intermediate_signing_key = SignedKey.model_validate_json(
        data.intermediateSigningKey.signedKey
    )
    if int(time.time() * 1000) > int(intermediate_signing_key.keyExpiration):
        raise ValueError("Expired intermediate signing key")

    # check signed message's signature
    # https://developers.google.com/wallet/generic/use-cases/use-callbacks-for-saves-and-deletions#verify-the-signature
    intermediate_public_key = _load_public_key(intermediate_signing_key.keyValue)
    signature = base64.decodebytes(bytes(data.signature, "utf-8"))
    signed_data = _construct_signed_data(
        "GooglePayWallet",
        issuer_id,
        PROTOCOL_VERSION,
        data.signedMessage,
    )
    try:
        intermediate_public_key.verify(signature, signed_data, ALGORITHM)
    except (ValueError, InvalidSignature):
        raise ValueError("Invalid signature")

    return message



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from .enums import BarcodeRenderEncoding
from .enums import BarcodeType
from .enums import TotpAlgorithm
from .localized_string import LocalizedString


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Barcode(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Barcode
    """

    # inherits kind (deprecated)
    type: BarcodeType = BarcodeType.BARCODE_TYPE_UNSPECIFIED
    renderEncoding: BarcodeRenderEncoding = (
        BarcodeRenderEncoding.RENDER_ENCODING_UNSPECIFIED
    )
    value: str | None = None
    alternateText: str | None = None
    showCodeText: LocalizedString | None = None


class TotpParameters(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/RotatingBarcode#totpparameters
    """

    key: str
    valueLength: int


class TotpDetails(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/RotatingBarcode#totpdetails
    """

    periodMillis: str
    algorithm: TotpAlgorithm = TotpAlgorithm.TOTP_ALGORITHM_UNSPECIFIED
    parameters: list[TotpParameters]


class RotatingBarcode(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/RotatingBarcode
    """

    type: BarcodeType = BarcodeType.BARCODE_TYPE_UNSPECIFIED
    renderEncoding: BarcodeRenderEncoding = (
        BarcodeRenderEncoding.RENDER_ENCODING_UNSPECIFIED
    )
    valuePattern: str | None = None
    totpDetails: TotpDetails | None = None
    alternateText: str | None = None
    showCodeText: LocalizedString | None = None
    # TODO initialRotatingBarcodeValues



from edutap.wallet_google.settings import ROOT_DIR


def test_session_manager_url(
    clean_registry_by_name, clean_registry_by_model
):  # noqa: F811
    from edutap.wallet_google.registry import register_model

    @register_model("Foo", url_part="foo")
    class Foo:
        pass

    from edutap.wallet_google.session import SessionManager

    manager = SessionManager()
    assert (
        manager.url("Foo")
        == "https://walletobjects.googleapis.com/walletobjects/v1/foo"
    )
    assert (
        manager.url("Foo", "/bar")
        == "https://walletobjects.googleapis.com/walletobjects/v1/foo/bar"
    )

    @register_model("Foo2", url_part="foobar")
    class Foo2:
        pass

    assert (
        manager.url("Foo2")
        == "https://walletobjects.googleapis.com/walletobjects/v1/foobar"
    )
    assert (
        manager.url("Foo2", "/baz")
        == "https://walletobjects.googleapis.com/walletobjects/v1/foobar/baz"
    )


def test_session_creation(monkeypatch):
    from edutap.wallet_google.session import SessionManager

    monkeypatch.setenv(
        "EDUTAP_WALLET_GOOGLE_CREDENTIALS_FILE",
        str(ROOT_DIR / "tests" / "data" / "credentials_fake.json"),
    )
    manager = SessionManager()
    session = manager.session
    assert session is not None
    assert manager.settings.credentials_file is not None
    assert session.credentials.scopes == [
        "https://www.googleapis.com/auth/wallet_object.issuer",
    ]

    from edutap.wallet_google.session import _THREADLOCAL

    assert _THREADLOCAL.session is session

    import threading

    def thread_check_different_session(other_session_id):
        session = manager.session  # noqa: F841
        assert id(_THREADLOCAL.session) != other_session_id

    threading.Thread(target=thread_check_different_session, args=[id(session)]).start()


def test_session_with_HTTPRecorder(tmp_path):
    from edutap.wallet_google.session import _THREADLOCAL
    from edutap.wallet_google.session import SessionManager

    delattr(_THREADLOCAL, "session")
    manager = SessionManager()
    manager.settings.record_api_calls_dir = tmp_path
    manager.settings.credentials_file = (
        ROOT_DIR / "tests" / "data" / "credentials_fake.json"
    )
    session = manager.session
    assert session.adapters["https://"].__class__.__name__ == "HTTPRecorder"



from pathlib import Path

import copy
import datetime
import json
import os
import pytest
import socket
import typing


DATA_PATH = Path(__file__).parent / "data"


@pytest.fixture
def clean_registry_by_name():
    """Fixture to provide a clean the google_wallet model registry."""
    from edutap.wallet_google.registry import _MODEL_REGISTRY_BY_NAME

    OLD_MODEL_REGISTRY_BY_NAME = copy.deepcopy(_MODEL_REGISTRY_BY_NAME)
    _MODEL_REGISTRY_BY_NAME.clear()
    yield _MODEL_REGISTRY_BY_NAME
    _MODEL_REGISTRY_BY_NAME.clear()
    _MODEL_REGISTRY_BY_NAME.update(OLD_MODEL_REGISTRY_BY_NAME)


@pytest.fixture
def clean_registry_by_model():
    """Fixture to provide a clean the google_wallet model registry."""
    from edutap.wallet_google.registry import _MODEL_REGISTRY_BY_MODEL

    OLD_MODEL_REGISTRY_BY_MODEL = copy.deepcopy(_MODEL_REGISTRY_BY_MODEL)
    _MODEL_REGISTRY_BY_MODEL.clear()
    yield _MODEL_REGISTRY_BY_MODEL
    _MODEL_REGISTRY_BY_MODEL.clear()
    _MODEL_REGISTRY_BY_MODEL.update(OLD_MODEL_REGISTRY_BY_MODEL)


@pytest.fixture
def mock_session(monkeypatch, requests_mock):
    """Fixture to provide a mock Google Wallet API session."""
    from edutap.wallet_google.session import _THREADLOCAL
    from edutap.wallet_google.session import SessionManager

    import requests

    _THREADLOCAL.session = None

    def mock_session(self):
        return requests.Session()

    monkeypatch.setattr(SessionManager, "session", property(mock_session))

    yield requests_mock

    _THREADLOCAL.session = None


@pytest.fixture
def mock_request_response(mock_session):
    """Fixture to load a mock request response from a json file.
    Prepares a mock response and status code for a given url and method.
    """

    def _load_mock_request_response(
        name: str, url: str, method: str, code=200
    ) -> dict[str, typing.Any]:
        data = {}
        for postfix in {"REQUEST", "RESPONSE"}:
            with open(DATA_PATH / f"{name}.REQUEST.json") as f:
                data[postfix.lower()] = json.load(f)
        mock_session.register_uri(
            method, url, json=data["response"]["body"], status_code=code
        )
        return data

    yield _load_mock_request_response


@pytest.fixture
def integration_test_id():
    prefix = os.environ.get(
        "EDUTAP_WALLET_GOOGLE_INTEGRATION_TEST_PREFIX", socket.gethostname()
    )
    timestamp = datetime.datetime.now(datetime.timezone.utc).strftime(
        "%Y-%m-%d_%H-%M-%S_%f"
    )
    yield f"{prefix}.{timestamp}"


@pytest.fixture
def mock_settings():
    from edutap.wallet_google.session import session_manager

    yield session_manager.settings
    del session_manager._settings


@pytest.fixture
def mock_fernet_encryption_key(mock_settings):
    mock_settings.fernet_encryption_key = "TDTPJVv24gha-jRX0apPgPpMDN2wX1kVSNNZdWXcz8E="


@pytest.fixture
def test_issuer_id():
    from edutap.wallet_google.session import session_manager

    yield session_manager.settings.test_issuer_id



from edutap.wallet_google.models.handlers import ImageData

import asyncio


class TestImageProvider:
    """
    Implementation of edutap.wallet_google.protocols.ImageProvider
    """

    async def image_by_id(self, image_id: str) -> ImageData:
        # return some predictable data for unit testing
        if image_id == "OK":
            return ImageData(mimetype="image/jpeg", data=b"mock-a-jepg")
        if image_id == "ERROR":
            raise LookupError("Image not found.")
        if image_id == "CANCEL":
            raise asyncio.CancelledError("Cancelled")
        if image_id == "TIMEOUT":
            await asyncio.sleep(0.5)
        raise Exception("Unexpected image_id")


class TestCallbackHandler:
    """
    Implementation of edutap.wallet_google.protocols.CallbackHandler

    Used in tests to simulate a callback handler and possible errors.
    """

    async def handle(
        self,
        class_id: str,
        object_id: str,
        event_type: str,
        exp_time_millis: int,
        count: int,
        nonce: str,
    ) -> None:
        if class_id.startswith("TIMEOUT"):
            await asyncio.sleep(exp_time_millis / 1000)
        elif nonce:
            return
        raise ValueError("test case errors if nonce is 0")



from ...registry import register_model
from ..datatypes.datetime import DateTime
from ..datatypes.enums import NotificationSettingsForUpdates
from ..datatypes.enums import RedemptionChannel
from ..datatypes.enums import ReviewStatus
from ..datatypes.general import GroupingInfo
from ..datatypes.general import Image
from ..datatypes.general import PassConstraints
from ..datatypes.general import Uri
from ..datatypes.localized_string import LocalizedString
from ..datatypes.location import LatLongPoint
from ..datatypes.loyalty import LoyaltyPoints
from ..datatypes.money import Money
from ..datatypes.retail import DiscoverableProgram
from ..datatypes.review import Review
from ..deprecated import DeprecatedAllowMultipleUsersPerObjectMixin
from ..deprecated import DeprecatedInfoModuleDataFieldMixin
from ..deprecated import DeprecatedKindFieldMixin
from ..deprecated import DeprecatedLocationsFieldMixin
from ..deprecated import DeprecatedVersionFieldMixin
from ..deprecated import DeprecatedWordMarkFieldMixin
from .bases import ClassModel
from .bases import ObjectModel
from .bases import StyleableMixin


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


@register_model("GiftCardClass", url_part="giftCardClass", plural="giftCardClasses")
class GiftCardClass(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/retail/gift-cards/rest/v1/giftcardclass
    """

    # inherits kind (deprecated)
    merchantName: str | None = None
    programLogo: Image | None = None
    pinLabel: str | None = None
    eventNumberLabel: str | None = None
    allowBarcodeRedemption: bool = False
    localizedMerchantName: LocalizedString | None = None
    localizedPinLabel: LocalizedString | None = None
    localizedEventNumberLabel: LocalizedString | None = None
    cardNumberLabel: str | None = None
    localizedCardNumberLabel: LocalizedString | None = None
    # inherits classTemplateInfo
    # inherits id
    # inherits version (deprecated)
    issuerName: str
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    homepageUri: Uri | None = None
    # inherits locations (deprecated)
    reviewStatus: ReviewStatus = ReviewStatus.REVIEW_STATUS_UNSPECIFIED
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    heroImage: Image | None = None
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    # inherits viewUnlockRequirement
    wideProgramLogo: Image | None = None
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("GiftCardObject", url_part="giftCardObject")
class GiftCardObject(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedLocationsFieldMixin,
    ObjectModel,
):
    """
    see: https://developers.google.com/wallet/retail/gift-cards/rest/v1/giftcardobject
    """

    # inherits kind (deprecated)
    classReference: GiftCardClass | None = None
    cardNumber: str | None = None
    pin: str | None = None
    balance: Money | None = None
    balanceUpdateTime: DateTime | None = None
    eventNumber: str | None = None
    # inherits id
    # inherits classId
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    # inherits locations (deprecated)
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool = False
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    # inherits rotatingBarcode
    heroImage: Image | None = None
    # inherits groupingInfo
    # inherits passConstraints
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData


@register_model("LoyaltyClass", url_part="loyaltyClass", plural="loyaltyClasses")
class LoyaltyClass(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass
    """

    programName: str | None = None
    programLogo: Image | None = None
    accountNameLabel: str | None = None
    accountIdLabel: str | None = None
    rewardsTierLabel: str | None = None
    rewardsTier: str | None = None
    localizedProgramName: LocalizedString | None = None
    localizedAccountNameLabel: LocalizedString | None = None
    localizedAccountIdLabel: LocalizedString | None = None
    localizedRewardsTierLabel: LocalizedString | None = None
    localizedRewardsTier: LocalizedString | None = None
    secondaryRewardsTierLabel: str | None = None
    localizedSecondaryRewardsTierLabel: LocalizedString | None = None
    secondaryRewardsTier: str | None = None
    localizedSecondaryRewardsTier: LocalizedString | None = None
    discoverableProgram: DiscoverableProgram | None = None
    # inherits classTemplateInfo
    # inherits id
    # inherits version (deprecated)
    issuerName: str | None = None
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    homepageUri: Uri | None = None
    locations: list[LatLongPoint] | None = None
    reviewStatus: ReviewStatus = ReviewStatus.REVIEW_STATUS_UNSPECIFIED
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    heroImage: Image | None = None
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    # inherits viewUnlockRequirement
    wideProgramLogo: Image | None = None
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("LoyaltyObject", url_part="loyaltyObject", plural="loyaltyObjects")
class LoyaltyObject(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedLocationsFieldMixin,
    ObjectModel,
):
    """
    data-type,
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyobject
    """

    # inherits kind (deprecated)
    classReference: LoyaltyClass | None = None
    accountName: str | None = None
    accountId: str | None = None
    loyaltyPoints: LoyaltyPoints | None = None
    linkedOfferIds: list[str] | None = None
    secondaryLoyaltyPoints: LoyaltyPoints | None = None
    # inherits id
    # inherits classId
    # inherits messages
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    # inherits locations (deprecated)
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool | None = None
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    # inherits rotatingBarcode
    heroImage: Image | None = None
    # inherits groupingInfo
    # inherits passConstraints
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData


@register_model("OfferClass", url_part="offerClass", plural="offerClasses")
class OfferClass(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/retail/offers/rest/v1/offerclass
    """

    # inherits kind (deprecated)
    title: str | None = None
    redemptionChannel: RedemptionChannel = (
        RedemptionChannel.REDEMPTION_CHANNEL_UNSPECIFIED
    )
    provider: str | None = None
    titleImage: Image | None = None
    details: str | None = None
    finePrint: str | None = None
    helpUri: Uri | None = None
    localizedTitle: LocalizedString | None = None
    localizedProvider: LocalizedString | None = None
    localizedDetails: LocalizedString | None = None
    localizedFinePrint: LocalizedString | None = None
    shortTitle: str | None = None
    localizedShortTitle: LocalizedString | None = None
    # inherits classTemplateInfo
    # inherits id
    # inherits version (deprecated)
    issuerName: str | None = None
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    homepageUri: Uri | None = None
    # inherits locations (deprecated)
    reviewStatus: ReviewStatus = ReviewStatus.REVIEW_STATUS_UNSPECIFIED
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    heroImage: Image | None = None
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    # inherits viewUnlockRequirement
    wideTitleImage: Image | None = None
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("OfferObject", url_part="offerObject")
class OfferObject(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    StyleableMixin,
    ObjectModel,
):
    """
    see: https://developers.google.com/wallet/retail/offers/rest/v1/offerobject
    """

    # inherits kind (deprecated)
    classReference: OfferClass | None = None
    # inherits id
    # inherits classId
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    locations: list[LatLongPoint] | None = None
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool = False
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    # inherits rotatingBarcode
    heroImage: Image | None = None
    groupingInfo: GroupingInfo | None = None
    passConstraints: PassConstraints | None = None
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData



from .models.bases import Model
from .models.datatypes.general import Pagination
from .models.datatypes.jwt import JWTClaims
from .models.datatypes.jwt import JWTPayload
from .models.datatypes.jwt import Reference
from .models.datatypes.message import Message
from .models.misc import AddMessageRequest
from .models.passes.bases import ClassModel
from .models.passes.bases import ObjectModel
from .registry import lookup_metadata_by_model_instance
from .registry import lookup_metadata_by_model_type
from .registry import lookup_metadata_by_name
from .registry import lookup_model_by_name
from .registry import raise_when_operation_not_allowed
from .session import session_manager
from collections.abc import Generator
from google.auth import crypt
from google.auth import jwt
from pydantic import ValidationError

import datetime
import json
import logging
import typing


logger = logging.getLogger(__name__)


def _validate_data(model: type[Model], data: dict[str, typing.Any] | Model) -> Model:
    """Takes a model and data, validates it and convert to a json string.

    :param model:      Pydantic model class to use for validation.
    :param data:       Data to pass to the Google RESTful API.
                       Either a simple python data structure using built-ins,
                       or a Pydantic model instance.
    :return:           data as an instance of the given model.
    """
    if not isinstance(data, (Model)):
        return model.model_validate(data)
    if not isinstance(data, model):
        raise ValueError(
            f"Model of given data mismatches given name. Expected {model}, got {type(data)}."
        )
    return data


def _validate_data_and_convert_to_json(
    model: type[Model],
    data: dict[str, typing.Any] | Model,
    *,
    existing: bool = False,
    resource_id_key: str = "id",
) -> tuple[str, str]:
    """Takes a model and data, validates it and convert to a json string.

    :param model:      Pydantic model class to use for validation.
    :param data:       Data to pass to the Google RESTful API.
                       Either a simple python data structure using built-ins,
                       or a Pydantic model instance.
    :param existing:   If True, the data is expected to be an existing object (i.e. on update).
    :return:           Tuple of resource-id and JSON string.
    """
    verified_data = _validate_data(model, data)
    verified_json = verified_data.model_dump_json(
        exclude_none=not existing,  # exclude None values when we create something new
        exclude_unset=True,  # exclude unset values - this are values not set explicitly by the code
        by_alias=True,
    )
    identifier = getattr(verified_data, resource_id_key)
    # nice to have: if existing: check if this is a valid identifier using read
    return (identifier, verified_json)


class WalletException(Exception):
    pass


class ObjectAlreadyExistsException(WalletException):
    pass


class QuotaExceededException(WalletException):
    pass


def new(
    name: str,
    data: dict[str, typing.Any] = {},
):
    """
    Factors a new registered Google Wallet Model by name, based on the given data.

    :param name:       Registered name of the model to use
    :param data:       Data to initialize the model with.
                       A simple JSON compatible Python data structure using built-ins.
    :raises Exception: When the data does not validate.
    :return:           The created model instance.
    """
    model = lookup_model_by_name(name)
    return _validate_data(model, data)


def create(
    data: Model,
) -> Model:
    """
    Creates a Google Wallet items. `C` in CRUD.

    :param data:                          Data to pass to the Google RESTful API.
                                          A model instance, has to be a registered model.
    :raises QuotaExceededException:       When the quota was exceeded.
    :raises ObjectAlreadyExistsException: When the id to be created already exists at Google.
    :raises WalletException:              When the response status code is not 200.
    :return:                              The created model based on the data returned by the Restful API.
    """
    model_metadata = lookup_metadata_by_model_instance(data)
    name = model_metadata["name"]
    raise_when_operation_not_allowed(name, "create")
    model = model_metadata["model"]
    resource_id, verified_json = _validate_data_and_convert_to_json(model, data)
    session = session_manager.session
    url = session_manager.url(name)
    headers = {"Content-Type": "application/json"}
    response = session.post(
        url=url,
        data=verified_json.encode("utf-8"),
        headers=headers,
    )
    if response.status_code == 403:
        raise QuotaExceededException(
            f"Quota exceeded while trying to create {name} {getattr(data, 'id', 'No ID')}"
        )
    elif response.status_code == 409:
        raise ObjectAlreadyExistsException(
            f"{name} {getattr(data, 'id', 'No ID')} already exists\n{response.text}"
        )
    elif response.status_code != 200:
        raise WalletException(
            f"Error on create of {name} {getattr(data, 'id', 'No ID')} at {url}: {response.status_code} - {response.text}"
        )

    logger.debug(f"RAW-Response: {response.content!r}")
    try:
        return model.model_validate_json(response.content)
    except ValidationError as e:
        logger.error(f"Validation Error: {e.errors()}")
        raise


def read(
    name: str,
    resource_id: str,
) -> Model:
    """
    Reads a Google Wallet Class or Object. `R` in CRUD.

    :param name:                      Registered name of the model to use
    :param resource_id:               Identifier of the resource to read from the Google RESTful API
    :QuotaExceededException:          When the quota was exceeded.
    :raises LookupError:              When the resource was not found (404).
    :raises WalletException           When the response status code is not 200 or 404.
    :return:                          the created model based on the data returned by the Restful API
    """
    raise_when_operation_not_allowed(name, "read")
    session = session_manager.session
    url = session_manager.url(name, f"/{resource_id}")
    response = session.get(url=url)

    if response.status_code == 403:
        raise QuotaExceededException(
            f"Quota exceeded while trying to read {name} {resource_id}"
        )
    elif response.status_code == 404:
        raise LookupError(f"{url}: {name} not found")

    if response.status_code == 200:
        model = lookup_model_by_name(name)
        logger.debug(f"RAW-Response: {response.content!r}")
        return model.model_validate_json(response.content)

    raise WalletException(f"{url} {response.status_code} - {response.text}")


def update(
    data: Model,
    *,
    partial: bool = True,
) -> Model:
    """
    Updates a Google Wallet Class or Object. `U` in CRUD.

    :param data:                    Data to pass to the Google RESTful API.
                                    A model instance, has to be a registered model.
    :param partial:                 Whether a partial update is executed or a full replacement.
    :raises QuotaExceededException: When the quota was exceeded
    :raises LookupError:            When the resource was not found (404)
    :raises WalletException:        When the response status code is not 200 or 404
    :return:                        The created model based on the data returned by the Restful API
    """
    model_metadata = lookup_metadata_by_model_instance(data)
    name = model_metadata["name"]
    raise_when_operation_not_allowed(name, "update")
    model = model_metadata["model"]
    resource_id, verified_json = _validate_data_and_convert_to_json(
        model, data, existing=True, resource_id_key=model_metadata["resource_id"]
    )
    session = session_manager.session
    if partial:
        response = session.patch(
            url=session_manager.url(name, f"/{resource_id}"),
            data=verified_json.encode("utf-8"),
        )
    else:
        response = session.put(
            url=session_manager.url(name, f"/{resource_id}"),
            data=verified_json.encode("utf-8"),
        )
    logger.debug(verified_json.encode("utf-8"))
    if response.status_code == 403:
        raise QuotaExceededException(
            f"Quota exceeded while trying to read {name} {resource_id}"
        )
    elif response.status_code == 404:
        raise LookupError(
            f"Error 404, {name} {getattr(data, 'id', 'No ID')} not found: - {response.text}"
        )
    if response.status_code != 200:
        raise WalletException(f"Error: {response.status_code} - {response.text}")

    logger.debug(f"RAW-Response: {response.content!r}")
    return model.model_validate_json(response.content)


def message(
    name: str,
    resource_id: str,
    message: dict[str, typing.Any] | Message,
) -> Model:
    """Sends a message to a Google Wallet Class or Object.

    :param name:                      Registered name of the model to use
    :param resource_id:               Identifier of the resource to send to
    :raises QuotaExceededException:   When the quota was exceeded
    :raises LookupError:              When the resource was not found (404)
    :raises WalletException:          When the response status code is not 200 or 404
    :return:                          The created Model object as returned by the Restful API
    """
    raise_when_operation_not_allowed(name, "message")
    model = lookup_model_by_name(name)

    if not isinstance(message, Message):
        message_validated = Message.model_validate(message)
    else:
        message_validated = message

    add_message = AddMessageRequest(message=message_validated)
    verified_json = add_message.model_dump_json(
        exclude_none=True,
    )
    url = session_manager.url(name, f"/{resource_id}/addMessage")
    response = session_manager.session.post(url=url, data=verified_json.encode("utf-8"))

    if response.status_code == 403:
        raise QuotaExceededException(
            f"Quota exceeded while trying to read {name} {resource_id}"
        )
    elif response.status_code == 404:
        raise LookupError(f"Error 404, {name} not found: - {response.text}")

    elif response.status_code != 200:
        raise WalletException(f"Error: {response.status_code} - {response.text}")

    logger.debug(f"RAW-Response: {response.content!r}")
    response_data = json.loads(response.content)
    return model.model_validate(response_data.get("resource"))


def listing(
    name: str,
    *,
    resource_id: str | None = None,
    issuer_id: str | None = None,
    result_per_page: int = 0,
    next_page_token: str | None = None,
) -> Generator[Model | str, None, None]:
    """Lists wallet related resources.

    It is possible to list all classes of an issuer. Parameter 'name' has to end with 'Class',
    all objects of a registered object type by it's classes resource id,
    Parameter 'name' has to end with 'Object'.
    To get all issuers, parameter 'name' has to be 'Issuer' and no further parameters are allowed.

    :param name:                    Registered name to base the listing on.
    :param resource_id:             Id of the class to list objects of.
                                    Only for object listings`
                                    Mutually exclusive with issuer_id.
    :param issuer_id:               Identifier of the issuer to list classes of.
                                    Only for class listings.
                                    If no resource_id is given and issuer_id is None, it will
                                    be fetched from the environment variable EDUTAP_WALLET_GOOGLE_ISSUER_ID.
                                    Mutually exclusive with resource_id.
    :param result_per_page:         Number of results per page to fetch.
                                    If omitted all results will be fetched and provided by the generator.
    :param next_page_token:         Token to get the next page of results.
    :raises QuotaExceededException: When the quota was exceeded
    :raises ValueError:             When input was invalid.
    :raises LookupError:            When the resource was not found (404)
    :raises WalletException:        When the response status code is not 200 or 404
    :return:                        Generator of the data as model-instances based on the data returned by the
                                    Restful API. When result_per_page is given, the generator will return
                                    a next_page_token after the last model-instance result.
    """
    raise_when_operation_not_allowed(name, "list")
    if resource_id and issuer_id:
        raise ValueError("resource_id and issuer_id are mutually exclusive")

    model = lookup_model_by_name(name)  # early, also to test if name is registered

    params = {}
    is_pageable = False
    if name.endswith("Object"):
        is_pageable = True
        if not resource_id:
            raise ValueError("resource_id of a class must be given to list its objects")
        params["classId"] = resource_id
    elif name.endswith("Class"):
        if not issuer_id:
            raise ValueError("issuer_id must be given to list classes")
        is_pageable = True
        params["issuerId"] = issuer_id

    if is_pageable:
        if next_page_token:
            params["token"] = next_page_token
        if result_per_page:
            params["maxResults"] = f"{result_per_page}"
        else:
            # default to 100, but this might need adjustment
            params["maxResults"] = "100"

    url = session_manager.url(name)
    session = session_manager.session
    while True:
        response = session.get(url=url, params=params)
        if response.status_code == 404:
            raise LookupError(f"Error 404, {name} not found: - {response.text}")

        if response.status_code != 200:
            raise Exception(f"Error: {response.status_code} - {response.text}")

        data = json.loads(response.content)
        for count, record in enumerate(data["resources"]):
            try:
                yield model.model_validate(record)
            except Exception:
                logger.exception(f"Error validating record {count}:\n{record}")
                raise
        if not is_pageable:
            break
        pagination = Pagination.model_validate(data["pagination"])
        if result_per_page > 0:
            if pagination.nextPageToken:
                yield pagination.nextPageToken
                break
        else:
            if pagination.nextPageToken:
                params["token"] = pagination.nextPageToken
                continue
        break


def _create_payload(models: list[ClassModel | ObjectModel | Reference]) -> JWTPayload:
    """Creates a payload for the JWT."""
    payload = JWTPayload()

    for model in models:
        if isinstance(model, Reference):
            if model.model_name is not None:
                name = lookup_metadata_by_name(model.model_name)["plural"]
            elif model.model_type is not None:
                name = lookup_metadata_by_model_type(model.model_type)["plural"]
        else:
            name = lookup_metadata_by_model_instance(model)["plural"]
        if getattr(payload, name) is None:
            setattr(payload, name, [])
        getattr(payload, name).append(model)
    return payload


def _convert_str_or_datetime_to_str(value: str | datetime.datetime) -> str:
    """convert and check the value to be a valid string for the JWT claim timestamps"""
    if isinstance(value, datetime.datetime):
        return str(int(value.timestamp()))
    if value == "":
        return value
    if not value.isdecimal():
        raise ValueError("string must be a decimal")
    if int(value) < 0:
        raise ValueError("string must be an int >= 0 number")
    if int(value) > 2**32:
        raise ValueError("string must be an int < 2**32 number")
    return value


def _create_claims(
    issuer: str,
    origins: list[str],
    models: list[ClassModel | ObjectModel | Reference],
    iat: str | datetime.datetime,
    exp: str | datetime.datetime,
) -> JWTClaims:
    """Creates a JWTClaims instance based on the given issuer, origins and models."""
    return JWTClaims(
        iss=issuer,
        iat=_convert_str_or_datetime_to_str(iat),
        exp=_convert_str_or_datetime_to_str(exp),
        origins=origins,
        payload=_create_payload(models),
    )


def save_link(
    models: list[ClassModel | ObjectModel | Reference],
    *,
    origins: list[str] = [],
    iat: str | datetime.datetime = "",
    exp: str | datetime.datetime = "",
) -> str:
    """
    Creates a link to save a Google Wallet Object to the wallet on the device.

    Besides the capability to save an object to the wallet, it is also able create classes on-the-fly.

    More information about the construction of the save_link can be found here:

    - https://developers.google.com/wallet/reference/rest/v1/jwt
    - https://developers.google.com/wallet/generic/web
    - https://developers.google.com/wallet/generic/use-cases/jwt

    :param models:      List of ObjectModels or ClassModels to save.
                        A resource can be an ObjectReference instance too.
    :param origins:     List of domains to approve for JWT saving functionality.
                        The Google Wallet API button will not render when the origins field is not defined.
                        You could potentially get an "Load denied by X-Frame-Options" or "Refused to display"
                        messages in the browser console when the origins field is not defined.
    :param: iat:        Issued At Time. The time when the JWT was issued.
    :param: exp:        Expiration Time. The time when the JWT expires.
    :return:            Link with JWT to save the resources to the wallet.
    """
    claims = _create_claims(
        session_manager.settings.credentials_info["client_email"],
        origins,
        models,
        iat=iat,
        exp=exp,
    )
    logger.debug(
        claims.model_dump_json(
            indent=2,
            exclude_unset=False,
            exclude_defaults=False,
            exclude_none=True,
        )
    )

    signer = crypt.RSASigner.from_service_account_file(
        session_manager.settings.credentials_file
    )
    jwt_string = jwt.encode(
        signer,
        claims.model_dump(
            mode="json",
            exclude_unset=False,
            exclude_defaults=False,
            exclude_none=True,
        ),
    ).decode("utf-8")
    logger.debug(jwt_string)
    if (jwt_len := len(jwt_string)) >= 1800:
        logger.debug(
            f"JWT-Length: {jwt_len} is larger than recommended 1800 bytes",
        )
    return f"{session_manager.settings.save_url}/{jwt_string}"



"""
see https://developers.google.com/wallet/generic/use-cases/use-callbacks-for-saves-and-deletions
"""

from .datatypes.enums import CamelCaseAliasEnum
from pydantic import BaseModel


# image fetching from data provider


class ImageData(BaseModel):
    mimetype: str
    data: bytes


#  Callback


class EventType(CamelCaseAliasEnum):
    SAVE = "SAVE"
    DEL = "DEL"


class SignedKey(BaseModel):
    keyValue: str
    keyExpiration: int


class IntermediateSigningKey(BaseModel):
    signedKey: str
    signatures: list[str]


class SignedMessage(BaseModel):
    classId: str
    objectId: str
    eventType: EventType
    expTimeMillis: int
    count: int
    nonce: str


class CallbackData(BaseModel):
    signature: str
    intermediateSigningKey: IntermediateSigningKey
    protocolVersion: str
    signedMessage: str  # Google sends this as a string, this get verified and the returned as a SignedMessage


# keys for callback data validation


class RootSigningPublicKey(BaseModel):
    """
    see https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#root-signing-keys
    """

    keyValue: str
    protocolVersion: str
    keyExpiration: str | None = None


class RootSigningPublicKeys(BaseModel):
    """
    see https://developers.google.com/pay/api/android/guides/resources/payment-data-cryptography#root-signing-keys
    """

    keys: list[RootSigningPublicKey]



from ..bases import Model
from .enums import RetailState
from .enums import SharedDataType
from .general import Uri


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class DiscoverableProgramMerchantSignupInfo(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass#discoverableprogrammerchantsignupinfo
    """

    signupWebsite: Uri
    signupSharedDatas: list[SharedDataType]


class DiscoverableProgramMerchantSigninInfo(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass#discoverableprogrammerchantsignininfo
    """

    signinWebsite: Uri


class DiscoverableProgram(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass#discoverableprogram
    """

    merchantSignupInfo: DiscoverableProgramMerchantSignupInfo | None = None
    merchantSigninInfo: DiscoverableProgramMerchantSigninInfo | None = None
    state: RetailState = RetailState.STATE_UNSPECIFIED



from ..bases import Model
from ..datatypes.enums import ActivationState
from ..datatypes.money import Money
from .enums import FareClass
from .general import LocalizedString
from pydantic import AnyHttpUrl
from pydantic import model_validator


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class ActivationOptions(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitclass#activationoptions
    """

    activationUrl: AnyHttpUrl | None = None
    allowReactivation: bool = False


class ActivationStatus(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#activationstatus
    """

    state: ActivationState = ActivationState.UNKNOWN_STATE


class TicketRestrictions(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#ticketrestrictions
    """

    routeRestrictions: LocalizedString | None = None
    routeRestrictionsDetails: LocalizedString | None = None
    timeRestrictions: LocalizedString | None = None
    otherRestrictions: LocalizedString | None = None


class TicketCost(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#ticketcost
    """

    faceValue: Money | None = None
    purchasePrice: Money | None = None
    discountMessage: LocalizedString | None = None


class TicketSeat(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#ticketseat
    """

    fareClass: FareClass = FareClass.FARE_CLASS_UNSPECIFIED
    customFareClass: LocalizedString | None = None
    coach: str | None = None
    seat: str | None = None
    seatAssignment: LocalizedString | None = None


class TicketLeg(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#ticketleg
    """

    originStationCode: str | None = None
    originName: LocalizedString | None = None
    destinationStationCode: str | None = None
    destinationName: LocalizedString | None = None
    departureDateTime: str | None = None
    arrivalDateTime: str | None = None
    fareName: LocalizedString | None = None
    carriage: str | None = None
    platform: str | None = None
    zone: str | None = None
    ticketSeat: TicketSeat | None = None
    ticketSeats: list[TicketSeat] | None = None
    transitOperatorName: LocalizedString | None = None
    transitTerminusName: LocalizedString | None = None

    @model_validator(mode="after")
    def check_one_of(self) -> "TicketLeg":
        if self.ticketSeat is None and self.ticketSeats is None:
            return self
        if (
            self.ticketSeat is not None
            and self.ticketSeats
            and len(self.ticketSeats) > 0
        ):
            raise ValueError("Only one of ticketSeat or ticketSeats must be set")
        if self.ticketSeats and len(self.ticketSeats) == 1:
            raise ValueError(
                "If only one seat is to be specified then use the ticketSeat field instead."
            )
        return self


class PurchaseDetails(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#purchasedetails
    """

    purchaseReceiptNumber: str | None = None
    purchaseDateTime: str | None = None
    accountId: str | None = None
    confirmationCode: str | None = None
    ticketCost: TicketCost | None = None


class DeviceContext(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#devicecontext
    """

    deviceToken: str | None = None



from edutap.wallet_google.models.handlers import ImageData

import pytest


def test_get_image_providers():
    from edutap.wallet_google.plugins import get_image_providers
    from edutap.wallet_google.protocols import ImageProvider

    plugins = get_image_providers()
    assert len(plugins) == 1
    assert isinstance(plugins[0], ImageProvider)


def test_get_callback_handlers():
    from edutap.wallet_google.plugins import get_callback_handlers
    from edutap.wallet_google.protocols import CallbackHandler

    plugins = get_callback_handlers()
    assert len(plugins) == 1
    assert isinstance(plugins[0], CallbackHandler)


def test_get_callback_handlers_empty(monkeypatch):
    from edutap.wallet_google.plugins import get_callback_handlers

    monkeypatch.setattr(
        "edutap.wallet_google.plugins.entry_points",
        lambda *args, **kw: [],
    )
    with pytest.raises(NotImplementedError):
        get_callback_handlers()


def test_get_image_providers_empty(monkeypatch):
    from edutap.wallet_google.plugins import get_image_providers

    monkeypatch.setattr(
        "edutap.wallet_google.plugins.entry_points",
        lambda *args, **kw: [],
    )
    with pytest.raises(NotImplementedError):
        get_image_providers()


def test_get_callback_handlers_wrong_type(monkeypatch):
    from edutap.wallet_google.plugins import get_callback_handlers

    class MockEP:
        name = "CallbackHandler"

        def load(self):
            return self

    monkeypatch.setattr(
        "edutap.wallet_google.plugins.entry_points",
        lambda *args, **kw: [MockEP()],
    )
    with pytest.raises(ValueError):
        get_callback_handlers()


def test_get_image_providers_wrong_type(monkeypatch):
    from edutap.wallet_google.plugins import get_image_providers

    class MockEP:
        name = "ImageProvider"

        def load(self):
            return self

    monkeypatch.setattr(
        "edutap.wallet_google.plugins.entry_points",
        lambda *args, **kw: [MockEP()],
    )
    with pytest.raises(ValueError):
        get_image_providers()


class DummyImageProvider:
    async def image_by_id(self, image_id: str) -> ImageData:
        raise NotImplementedError


class DummyCallbackHandler:
    async def handle(
        self,
        class_id: str,
        object_id: str,
        event_type: str,
        exp_time_millis: int,
        count: int,
        nonce: str,
    ) -> None: ...


def test_add_plugin(monkeypatch):
    """
    test adding plugins at runtime
    """
    from edutap.wallet_google.plugins import add_plugin
    from edutap.wallet_google.plugins import get_callback_handlers
    from edutap.wallet_google.plugins import get_image_providers

    count_image_providers = len(get_image_providers())
    count_callback_handlers = len(get_callback_handlers())

    add_plugin("ImageProvider", DummyImageProvider)
    add_plugin("CallbackHandler", DummyCallbackHandler)

    with pytest.raises(TypeError):
        add_plugin("CallbackHandler", DummyImageProvider)

    plugins = get_image_providers()
    assert len(plugins) == 1 + count_image_providers
    assert isinstance(plugins[-1], DummyImageProvider)

    plugins = get_callback_handlers()
    assert len(plugins) == 1 + count_callback_handlers
    assert isinstance(plugins[-1], DummyCallbackHandler)



from ...registry import register_model
from ..datatypes.enums import GenericType
from ..datatypes.general import GroupingInfo
from ..datatypes.localized_string import LocalizedString
from ..datatypes.notification import Notifications
from .bases import ClassModel
from .bases import CommonLogosMixin
from .bases import ObjectModel
from .bases import StyleableMixin


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


@register_model(
    "GenericClass",
    url_part="genericClass",
    plural="genericClasses",
    can_message=True,
)
class GenericClass(ClassModel):
    """
    The GenericClass is the implicitly the base class for all other Wallet Class models.
    The Google documentation does not mention this fact.
    This might change in future updates, do not depend on this assumption!

    For now, technically, here are no extra attributes defined!
    In code the ClassModel is the base class.

    see: https://developers.google.com/wallet/generic/rest/v1/genericclass
    """

    # inherits id
    # inherits classTemplateInfo
    # inherits infoModuleData
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits enableSmartTap
    # inherits redemptionIssuers
    # inherits securityAnimation
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits viewUnlockRequirement
    # inherits messages
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("GenericObject", url_part="genericObject", can_message=True)
class GenericObject(StyleableMixin, CommonLogosMixin, ObjectModel):
    """
    The GenericObject is a specific object and does not act as the base for other wallet objects!

    see: https://developers.google.com/wallet/generic/rest/v1/genericobject
    """

    genericType: GenericType = GenericType.GENERIC_TYPE_UNSPECIFIED
    cardTitle: LocalizedString | None = None
    subheader: LocalizedString | None = None
    header: LocalizedString | None = None
    # inherits logo
    # inherits hexBackgroundColor
    notifications: Notifications | None = None
    # inherits id
    # inherits classId
    # inherits barcode
    # inherits heroImage
    # inherits validTimeInterval
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    groupingInfo: GroupingInfo | None = None
    # inherits smartTapRedemptionValue
    # inherits rotatingBarcode
    # inherits state
    # inherits hasUsers
    # inherits messages
    # inherits passConstraints
    # inherits wideLogo
    # inherits saveRestrictions
    # inherits valueAddedModuleData
    # inherits linkedObjectIds



from .bases import ClassModel  # noqa: F401
from .bases import ObjectModel  # noqa: F401
from .generic import GenericClass  # noqa: F401
from .generic import GenericObject  # noqa: F401
from .retail import GiftCardClass  # noqa: F401
from .retail import GiftCardObject  # noqa: F401
from .retail import LoyaltyClass  # noqa: F401
from .retail import LoyaltyObject  # noqa: F401
from .retail import OfferClass  # noqa: F401
from .retail import OfferObject  # noqa: F401
from .tickets_and_transit import EventTicketClass  # noqa: F401
from .tickets_and_transit import EventTicketObject  # noqa: F401
from .tickets_and_transit import FlightClass  # noqa: F401
from .tickets_and_transit import FlightObject  # noqa: F401
from .tickets_and_transit import TransitClass  # noqa: F401
from .tickets_and_transit import TransitObject  # noqa: F401



from ..bases import Model
from .enums import DateFormat
from .enums import PredefinedItem
from .enums import TransitOption
from pydantic import Field
from typing import Annotated
from typing_extensions import deprecated


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class FieldReference(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#fieldreference
    """

    fieldPath: str
    dateFormat: DateFormat | None = None


class FieldSelector(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#fieldselector
    """

    fields: list[FieldReference]


class TemplateItem(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#templateitem
    """

    firstValue: FieldSelector | None = None
    secondValue: FieldSelector | None = None
    predefinedItem: PredefinedItem | None = None


class BarcodeSectionDetail(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#barcodesectiondetail
    """

    fieldSelector: FieldSelector


class CardBarcodeSectionDetails(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardbarcodesectiondetails
    """

    firstTopDetail: BarcodeSectionDetail | None = None
    firstBottomDetail: BarcodeSectionDetail | None = None
    secondTopDetail: BarcodeSectionDetail | None = None


class CardRowOneItem(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardrowoneitem
    """

    item: TemplateItem | None = None


class CardRowTwoItems(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardrowtwoitems
    """

    startItem: TemplateItem | None = None
    endItem: TemplateItem | None = None


class CardRowThreeItems(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardrowthreeitems
    """

    startItem: TemplateItem | None = None
    middleItem: TemplateItem | None = None
    endItem: TemplateItem | None = None


class CardRowTemplateInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardrowtemplateinfo
    """

    oneItem: CardRowOneItem | None = None
    twoItems: CardRowTwoItems | None = None
    threeItems: CardRowThreeItems | None = None


class CardTemplateOverride(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#cardtemplateoverride
    """

    cardRowTemplateInfos: list[CardRowTemplateInfo] | None = None


class DetailsItemInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#detailsiteminfo
    """

    item: TemplateItem | None = None


class DetailsTemplateOverride(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#detailstemplateoverride
    """

    detailsItemInfos: list[DetailsItemInfo] | None = None


class FirstRowOption(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#firstrowoption
    """

    transitOption: TransitOption | None = None
    fieldOption: FieldSelector | None = None


class ListTemplateOverride(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#listtemplateoverride
    """

    firstRowOption: FirstRowOption | None = None
    secondRowOption: FieldSelector | None = None
    thirdRowOption: Annotated[
        FieldSelector | None,
        Field(
            deprecated=deprecated(
                'The Attribute "thirdRowOption" is deprecated on "ListTemplateOverride". Setting it will have no effect on what the user sees.'
            ),
            default=None,
            exclude=True,
        ),
    ]


class ClassTemplateInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo
    """

    cardBarcodeSectionDetails: CardBarcodeSectionDetails | None = None
    cardTemplateOverride: CardTemplateOverride | None = None
    detailsTemplateOverride: DetailsTemplateOverride | None = None
    listTemplateOverride: ListTemplateOverride | None = None


# classTemplateInfo/listTemplateOverride/listTemplateOverride/firstRowOption/fieldOption/fields[]/[fieldPath|dateFormat]



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Money(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/Money
    """

    # inherits kind (deprecated)
    micros: str | None = None
    currencyCode: str | None = None



from .protocols import CallbackHandler
from .protocols import ImageProvider
from importlib.metadata import entry_points

import typing


_POSSIBLE_PLUGINS = {
    "ImageProvider": ImageProvider,
    "CallbackHandler": CallbackHandler,
}

_PLUGIN_REGISTRY: dict[str, list[CallbackHandler | ImageProvider]] = {
    "ImageProvider": [],
    "CallbackHandler": [],
}


def add_plugin(name: str, klass: CallbackHandler | ImageProvider):
    if not isinstance(klass, _POSSIBLE_PLUGINS[name]):
        raise TypeError(f"{klass} not implements {name}")
    _PLUGIN_REGISTRY[name].append(klass)


def get_plugins(name: str) -> list[CallbackHandler | ImageProvider]:
    eps = entry_points(group="edutap.wallet_google.plugins")
    plugins = [ep.load() for ep in eps if ep.name.startswith(name)]
    plugins += _PLUGIN_REGISTRY.get(name, [])
    if not plugins:
        raise NotImplementedError("No image provider plug-in found")
    for plugin in plugins:
        if not isinstance(plugin, _POSSIBLE_PLUGINS[name]):
            raise ValueError(f"{plugin} not implements {name}")
    return [plugin() for plugin in plugins]


def get_image_providers() -> list[ImageProvider]:
    return typing.cast(list[ImageProvider], get_plugins("ImageProvider"))


def get_callback_handlers() -> list[CallbackHandler]:
    return typing.cast(list[CallbackHandler], get_plugins("CallbackHandler"))



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin

# TODO: imports itself, i.e. shadows the datetime module
import datetime


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class DateTime(Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/DateTime
    """

    date: datetime.datetime


class TimeInterval(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/TimeInterval
    """

    # inherits kind (deprecated)
    start: DateTime | None = None
    end: DateTime | None = None



from ..bases import CamelCaseAliasEnum


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Action(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/smarttap#action
    """

    ACTION_UNSPECIFIED = "ACTION_UNSPECIFIED"
    S2AP = "S2AP"
    SIGN_UP = "SIGN_UP"


class ActivationState(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#state
    """

    UNKNOWN_STATE = "UNKNOWN_STATE"
    NOT_ACTIVATED = "NOT_ACTIVATED"
    ACTIVATED = "ACTIVATED"


class AnimationType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/SecurityAnimation#animationtype
    """

    ANIMATION_UNSPECIFIED = "ANIMATION_UNSPECIFIED"
    FOIL_SHIMMER = "FOIL_SHIMMER"


class BarcodeRenderEncoding(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/BarcodeRenderEncoding
    """

    RENDER_ENCODING_UNSPECIFIED = "RENDER_ENCODING_UNSPECIFIED"
    UTF_8 = "UTF_8"


class BarcodeType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/BarcodeType
    """

    BARCODE_TYPE_UNSPECIFIED = "BARCODE_TYPE_UNSPECIFIED"
    AZTEC = "AZTEC"
    CODE_39 = "CODE_39"
    CODE_128 = "CODE_128"
    CODABAR = "CODABAR"
    DATA_MATRIX = "DATA_MATRIX"
    EAN_8 = "EAN_8"
    EAN_13 = "EAN_13"
    ITF_14 = "ITF_14"
    PDF_417 = "PDF_417"
    QR_CODE = "QR_CODE"
    UPC_A = "UPC_A"
    TEXT_ONLY = "TEXT_ONLY"


class BoardingPolicy(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#boardingpolicy
    """

    BOARDING_POLICY_UNSPECIFIED = "BOARDING_POLICY_UNSPECIFIED"
    ZONE_BASED = "ZONE_BASED"
    GROUP_BASED = "GROUP_BASED"
    BOARDING_POLICY_OTHER = "BOARDING_POLICY_OTHER"


class ConcessionCategory(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#concessioncategory
    """

    CONCESSION_CATEGORY_UNSPECIFIED = "CONCESSION_CATEGORY_UNSPECIFIED"
    ADULT = "ADULT"
    CHILD = "CHILD"
    SENIOR = "SENIOR"


class ConfirmationCodeLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#confirmationcodelabel
    """

    CONFIRMATION_CODE_LABEL_UNSPECIFIED = "CONFIRMATION_CODE_LABEL_UNSPECIFIED"
    CONFIRMATION_CODE = "CONFIRMATION_CODE"
    CONFIRMATION_NUMBER = "CONFIRMATION_NUMBER"
    ORDER_NUMBER = "ORDER_NUMBER"
    RESERVATION_NUMBER = "RESERVATION_NUMBER"


class DateFormat(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/ClassTemplateInfo#dateformat
    """

    DATE_FORMAT_UNSPECIFIED = "DATE_FORMAT_UNSPECIFIED"
    DATE_TIME = "DATE_TIME"
    DATE_ONLY = "DATE_ONLY"
    TIME_ONLY = "TIME_ONLY"
    DATE_TIME_YEAR = "DATE_TIME_YEAR"
    DATE_YEAR = "DATE_YEAR"


class DoorsOpenLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#doorsopenlabel
    """

    DOORS_OPEN_LABEL_UNSPECIFIED = "DOORS_OPEN_LABEL_UNSPECIFIED"
    DOORS_OPEN = "DOORS_OPEN"
    GATES_OPEN = "GATES_OPEN"


class FareClass(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#fareclass
    """

    FARE_CLASS_UNSPECIFIED = "FARE_CLASS_UNSPECIFIED"
    ECONOMY = "ECONOMY"
    FIRST = "FIRST"
    BUSINESS = "BUSINESS"


class FlightStatus(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#flightstatus
    """

    FLIGHT_STATUS_UNSPECIFIED = "FLIGHT_STATUS_UNSPECIFIED"
    SCHEDULED = "SCHEDULED"
    ACTIVE = "ACTIVE"
    LANDED = "LANDED"
    CANCELLED = "CANCELLED"
    REDIRECTED = "REDIRECTED"
    DIVERTED = "DIVERTED"


class GateLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#gatelabel
    """

    GATE_LABEL_UNSPECIFIED = "GATE_LABEL_UNSPECIFIED"
    GATE = "GATE"
    DOOR = "DOOR"
    ENTRANCE = "ENTRANCE"


class GenericType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/genericobject#generictype
    """

    GENERIC_TYPE_UNSPECIFIED = "GENERIC_TYPE_UNSPECIFIED"  # Unspecified generic type.
    GENERIC_SEASON_PASS = "GENERIC_SEASON_PASS"  # Season pass
    GENERIC_UTILITY_BILLS = "GENERIC_UTILITY_BILLS"  # Utility bills
    GENERIC_PARKING_PASS = "GENERIC_PARKING_PASS"  # Parking pass
    GENERIC_VOUCHER = "GENERIC_VOUCHER"  # Voucher
    GENERIC_GYM_MEMBERSHIP = "GENERIC_GYM_MEMBERSHIP"  # Gym membership cards
    GENERIC_LIBRARY_MEMBERSHIP = (
        "GENERIC_LIBRARY_MEMBERSHIP"  # Library membership cards
    )
    GENERIC_RESERVATIONS = "GENERIC_RESERVATIONS"  # Reservations
    GENERIC_AUTO_INSURANCE = "GENERIC_AUTO_INSURANCE"  # Auto-insurance cards
    GENERIC_HOME_INSURANCE = "GENERIC_HOME_INSURANCE"  # Home-insurance cards
    GENERIC_ENTRY_TICKET = "GENERIC_ENTRY_TICKET"  # Entry tickets
    GENERIC_RECEIPT = "GENERIC_RECEIPT"  # Receipts
    GENERIC_OTHER = "GENERIC_OTHER"  # Other type


class MessageType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/Message#messagetype
    """

    MESSAGE_TYPE_UNSPECIFIED = "MESSAGE_TYPE_UNSPECIFIED"
    TEXT = "TEXT"
    TEXT_AND_NOTIFY = "TEXT_AND_NOTIFY"
    EXPIRATION_NOTIFICATION = "EXPIRATION_NOTIFICATION"


class MultipleDevicesAndHoldersAllowedStatus(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/MultipleDevicesAndHoldersAllowedStatus
    """

    STATUS_UNSPECIFIED = "STATUS_UNSPECIFIED"
    MULTIPLE_HOLDERS = "MULTIPLE_HOLDERS"
    ONE_USER_ALL_DEVICES = "ONE_USER_ALL_DEVICES"
    ONE_USER_ONE_DEVICE = "ONE_USER_ONE_DEVICE"


class NfcConstraint(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/PassConstraints#NfcConstraint
    """

    NFC_CONSTRAINT_UNSPECIFIED = "NFC_CONSTRAINT_UNSPECIFIED"
    BLOCK_PAYMENT = "BLOCK_PAYMENT"
    BLOCK_CLOSED_LOOP_TRANSIT = "BLOCK_CLOSED_LOOP_TRANSIT"


class NotificationSettingsForUpdates(CamelCaseAliasEnum):
    """
    see https://developers.google.com/wallet/reference/rest/v1/NotificationSettingsForUpdates
    """

    NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED = (
        "NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED"
    )
    NOTIFY_ON_UPDATE = "NOTIFY_ON_UPDATE"


class PassengerType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#passengertype
    """

    PASSENGER_TYPE_UNSPECIFIED = "PASSENGER_TYPE_UNSPECIFIED"
    SINGLE_PASSENGER = "SINGLE_PASSENGER"
    MULTIPLE_PASSENGERS = "MULTIPLE_PASSENGERS"


class PredefinedItem(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/retail/offers/rest/v1/ClassTemplateInfo#predefineditem
    """

    PREDEFINED_ITEM_UNSPECIFIED = "PREDEFINED_ITEM_UNSPECIFIED"
    FREQUENT_FLYER_PROGRAM_NAME_AND_NUMBER = "FREQUENT_FLYER_PROGRAM_NAME_AND_NUMBER"
    FLIGHT_NUMBER_AND_OPERATING_FLIGHT_NUMBER = (
        "FLIGHT_NUMBER_AND_OPERATING_FLIGHT_NUMBER"
    )


class RedemptionChannel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/retail/offers/rest/v1/offerclass#OfferClass.RedemptionChannel
    """

    REDEMPTION_CHANNEL_UNSPECIFIED = "REDEMPTION_CHANNEL_UNSPECIFIED"
    INSTORE = "INSTORE"
    ONLINE = "ONLINE"
    BOTH = "BOTH"
    TEMPORARY_PRICE_REDUCTION = "TEMPORARY_PRICE_REDUCTION"


class ReviewStatus(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/ReviewStatus
    """

    REVIEW_STATUS_UNSPECIFIED = "REVIEW_STATUS_UNSPECIFIED"
    UNDER_REVIEW = "UNDER_REVIEW"
    APPROVED = "APPROVED"
    REJECTED = "REJECTED"
    DRAFT = "DRAFT"


class Role(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/permissions#role
    """

    ROLE_UNSPECIFIED = "ROLE_UNSPECIFIED"
    OWNER = "OWNER"
    READER = "READER"
    WRITER = "WRITER"


class RowLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#rowlabel
    """

    ROW_LABEL_UNSPECIFIED = "ROW_LABEL_UNSPECIFIED"
    ROW = "ROW"


class ScreenshotEligibility(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/PassConstraints#screenshoteligibility
    """

    SCREENSHOT_ELIGIBILITY_UNSPECIFIED = "SCREENSHOT_ELIGIBILITY_UNSPECIFIED"
    ELIGIBLE = "ELIGIBLE"
    INELIGIBLE = "INELIGIBLE"


class SeatClassPolicy(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#seatclasspolicy
    """

    SEAT_CLASS_POLICY_UNSPECIFIED = "SEAT_CLASS_POLICY_UNSPECIFIED"
    CABIN_BASED = "CABIN_BASED"
    CLASS_BASED = "CLASS_BASED"
    TIER_BASED = "TIER_BASED"
    SEAT_CLASS_POLICY_OTHER = "SEAT_CLASS_POLICY_OTHER"


class SeatLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#seatlabel
    """

    SEAT_LABEL_UNSPECIFIED = "SEAT_LABEL_UNSPECIFIED"
    SEAT = "SEAT"


class SectionLabel(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#sectionlabel
    """

    SECTION_LABEL_UNSPECIFIED = "SECTION_LABEL_UNSPECIFIED"
    SECTION = "SECTION"
    THEATER = "THEATER"


class SharedDataType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass#shareddatatype
    """

    SHARED_DATA_TYPE_UNSPECIFIED = "SHARED_DATA_TYPE_UNSPECIFIED"
    FIRST_NAME = "FIRST_NAME"
    LAST_NAME = "LAST_NAME"
    STREET_ADDRESS = "STREET_ADDRESS"  # single line address field
    ADDRESS_LINE_1 = "ADDRESS_LINE_1"  # multi line address fields
    ADDRESS_LINE_2 = "ADDRESS_LINE_2"
    ADDRESS_LINE_3 = "ADDRESS_LINE_3"
    CITY = "CITY"
    STATE = "STATE"
    ZIPCODE = "ZIPCODE"
    COUNTRY = "COUNTRY"
    EMAIL = "EMAIL"
    PHONE = "PHONE"


class State(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/State
         https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/State
    """

    STATE_UNSPECIFIED = "STATE_UNSPECIFIED"
    ACTIVE = "ACTIVE"
    COMPLETED = "COMPLETED"
    EXPIRED = "EXPIRED"
    INACTIVE = "INACTIVE"


class RetailState(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyclass#state
    """

    STATE_UNSPECIFIED = "STATE_UNSPECIFIED"
    TRUSTED_TESTERS = "TRUSTED_TESTERS"
    LIVE = "LIVE"
    DISABLED = "DISABLED"


class TicketStatus(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#ticketstatus
    """

    TICKET_STATUS_UNSPECIFIED = "TICKET_STATUS_UNSPECIFIED"
    USED = "USED"
    REFUNDED = "REFUNDED"
    EXCHANGED = "EXCHANGED"


class TotpAlgorithm(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/RotatingBarcode#totpalgorithm
    """

    TOTP_ALGORITHM_UNSPECIFIED = "TOTP_ALGORITHM_UNSPECIFIED"
    TOTP_SHA1 = "TOTP_SHA1"


class TransitType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitclass#transittype
    """

    TRANSIT_TYPE_UNSPECIFIED = "TRANSIT_TYPE_UNSPECIFIED"
    BUS = "BUS"
    RAIL = "RAIL"
    TRAM = "TRAM"
    FERRY = "FERRY"
    OTHER = "OTHER"


class TransitOption(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ClassTemplateInfo#transitoption
    """

    TRANSIT_OPTION_UNSPECIFIED = "TRANSIT_OPTION_UNSPECIFIED"
    ORIGIN_AND_DESTINATION_NAMES = "ORIGIN_AND_DESTINATION_NAMES"
    ORIGIN_AND_DESTINATION_CODES = "ORIGIN_AND_DESTINATION_CODES"
    ORIGIN_NAME = "ORIGIN_NAME"


class TripType(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject#triptype
    """

    TRIP_TYPE_UNSPECIFIED = "TRIP_TYPE_UNSPECIFIED"
    ROUND_TRIP = "ROUND_TRIP"
    ONE_WAY = "ONE_WAY"


class ViewUnlockRequirement(CamelCaseAliasEnum):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ViewUnlockRequirement
    """

    VIEW_UNLOCK_REQUIREMENT_UNSPECIFIED = "VIEW_UNLOCK_REQUIREMENT_UNSPECIFIED"
    UNLOCK_NOT_REQUIRED = "UNLOCK_NOT_REQUIRED"
    UNLOCK_REQUIRED_TO_VIEW = "UNLOCK_REQUIRED_TO_VIEW"



from ..models.handlers import CallbackData
from ..plugins import get_callback_handlers
from ..plugins import get_image_providers
from ..session import session_manager
from ..utils import decrypt_data
from .validate import verified_signed_message
from fastapi import APIRouter
from fastapi import Request
from fastapi import Response
from fastapi.exceptions import HTTPException
from fastapi.logger import logger
from fastapi.responses import JSONResponse

import asyncio


# define routers for all use cases: callback, images, and the combined router (at bottom of file)
router_callback = APIRouter(
    prefix=session_manager.settings.handler_prefix_callback,
    tags=["edutap.wallet_google"],
)
router_images = APIRouter(
    prefix=session_manager.settings.handler_prefix_images,
    tags=["edutap.wallet_google"],
)


@router_callback.post("/callback")
async def handle_callback(request: Request, callback_data: CallbackData):
    """FastAPI handler for the callback endpoint.

    It is called by Google Wallet API when a user interacts with a pass.
    The callback is triggered on save and delete of a pass in the wallet.
    """
    # get the registered callback handlers
    try:
        handlers = get_callback_handlers()
    except NotImplementedError:
        raise HTTPException(
            status_code=500, detail="No callback handlers were registered."
        )

    # extract and verify message (given verification is not disabled)
    callback_message = verified_signed_message(callback_data)
    logger.debug(f"Got message {callback_message}")

    # call each handler asynchronously
    try:
        results: list = (
            # this could be replaced by async with asyncio.timeout(5.0) in Py 3.11
            # see also https://hynek.me/articles/waiting-in-asyncio/
            await asyncio.wait_for(
                asyncio.gather(
                    *(
                        handler.handle(
                            callback_message.classId,
                            callback_message.objectId,
                            callback_message.eventType.value,
                            callback_message.expTimeMillis,
                            callback_message.count,
                            callback_message.nonce,
                        )
                        for handler in handlers
                    ),
                    return_exceptions=True,
                ),
                timeout=session_manager.settings.handlers_callback_timeout,
            )
        )
    except asyncio.TimeoutError:
        logger.exception(
            f"Timeout after {session_manager.settings.handlers_callback_timeout}s while handling the callbacks.",
        )
        raise HTTPException(
            status_code=500, detail="Error while handling the callbacks (timeout)."
        )
    # results is a list of exceptions or None
    if any(results):
        logger.error("Error while handling a callbacks.")
        raise HTTPException(
            status_code=500, detail="Error while handling the callbacks (exception)."
        )
    return JSONResponse(content={"status": "success"})


@router_images.get("/images/{encrypted_image_id:path}")
async def handle_image(request: Request, encrypted_image_id: str):
    """FastAPI handler for the image endpoint.

    It is called by Google Wallet API to fetch images for a pass.
    """
    # get the registered image providers
    try:
        handlers = get_image_providers()
    except NotImplementedError:
        raise HTTPException(
            status_code=500, detail="No image providers were registered."
        )
    if len(handlers) > 1:
        logger.error("Multiple image providers found, abort.")
        raise HTTPException(
            status_code=500, detail="Multiple image providers found, abort."
        )

    handler = handlers[0]

    # decrypt the image id
    image_id = decrypt_data(encrypted_image_id)

    try:
        result = await asyncio.wait_for(
            handler.image_by_id(image_id),
            timeout=session_manager.settings.handlers_image_timeout,
        )
    except asyncio.TimeoutError:
        logger.exception(
            "Timeout Timeout after {session_manager.settings.handlers_image_timeout}s while handling the image.",
        )
        raise HTTPException(
            status_code=500, detail="Error while handling the image (timeout)."
        )
    except asyncio.CancelledError:
        logger.exception(
            "Cancelled while handling the image.",
        )
        raise HTTPException(
            status_code=500, detail="Error while handling the image (cancel)."
        )
    except LookupError:
        raise HTTPException(status_code=404, detail="Image not found.")
    except Exception:
        logger.exception(
            "Error while handling a image.",
        )
        raise HTTPException(
            status_code=500, detail="Error while handling the image (exception)."
        )
    return Response(
        content=result.data,
        media_type=result.mimetype,
        headers={"Cache-Control": session_manager.settings.handler_image_cache_control},
    )


# needs to be included after the routers are defined
router = APIRouter(
    prefix=session_manager.settings.handler_prefix,
)
router.include_router(router_callback)
router.include_router(router_images)



from .models.handlers import ImageData
from typing import Protocol
from typing import runtime_checkable


@runtime_checkable
class ImageProvider(Protocol):

    async def image_by_id(self, image_id: str) -> ImageData:
        """
        :param image_id: Unique image identifier as string.
        :return: ImageData instance.

        :raises: LookupError if image_id was not found.
        """


@runtime_checkable
class CallbackHandler(Protocol):

    async def handle(
        self,
        class_id: str,
        object_id: str,
        event_type: str,
        exp_time_millis: int,
        count: int,
        nonce: str,
    ) -> None:
        """
        :param class_id: ClassId
        :param object_id: ObjectId
        :param event_type: EventType
        :param exp_time_millis: Expiration time in milliseconds
        :param count: count retries
        :param nonce: nonce UID

        :raises: ValueError
        """



import pytest


def test_reference_validator_missing():
    from edutap.wallet_google.models.passes.bases import Reference

    import pydantic

    with pytest.raises(pydantic.ValidationError):
        Reference(id="1234")


def test_reference_validator_both():
    from edutap.wallet_google.models.passes.bases import Reference
    from edutap.wallet_google.models.passes.generic import GenericObject

    import pydantic

    with pytest.raises(pydantic.ValidationError):
        Reference(id="1234", model_name="GenericObject", model_type=GenericObject)


def test_loyality_LoyaltyPointsBalance_validator_OK():
    from edutap.wallet_google.models.datatypes.loyalty import LoyaltyPointsBalance
    from edutap.wallet_google.models.datatypes.money import Money

    LoyaltyPointsBalance(string="1234")
    LoyaltyPointsBalance(int=1234)
    LoyaltyPointsBalance(double=1234.00)
    LoyaltyPointsBalance(money=Money(micros="123400", currencyCode="USD"))


def test_loyality_LoyaltyPointsBalance_validator_failures():
    from edutap.wallet_google.models.datatypes.loyalty import LoyaltyPointsBalance
    from edutap.wallet_google.models.datatypes.money import Money

    import pydantic

    # at least one
    with pytest.raises(pydantic.ValidationError):
        LoyaltyPointsBalance()

    # not 2
    with pytest.raises(pydantic.ValidationError):
        LoyaltyPointsBalance(int=1234, string="1234")

    # not 3
    with pytest.raises(pydantic.ValidationError):
        LoyaltyPointsBalance(int=1234, string="1234", double=1234.00)

    # not 4
    with pytest.raises(pydantic.ValidationError):
        LoyaltyPointsBalance(
            int=1234,
            string="1234",
            double=1234.00,
            money=Money(
                micros="123400",
                currencyCode="USD",
            ),
        )



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class LatLongPoint(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/LatLongPoint
    """

    # inherits kind (deprecated)
    latitude: float = Field(ge=-90.0, le=90.0)
    longitude: float = Field(ge=-180.0, le=180.0)



"""
Many Google wallet models or datatypes have deprecated attributes.
Those must not be serialized again, nor repeated in the special models.
To handle this, we create mixins which can be added to the models which still have the deprecated fields.
"""

from .bases import Model
from pydantic import AnyUrl
from pydantic import Field
from typing import Annotated
from typing_extensions import deprecated


class DeprecatedKindFieldMixin:
    """
    Mixin to add the deprecated kind field to a model.
    May be removed in the future.
    """

    kind: Annotated[
        str,
        Field(
            deprecated=deprecated(
                'Attribute "kind" was used in the past to identify resources but is now deprecated.'
            ),
            exclude=True,
            default=None,
        ),
    ]


class DeprecatedAllowMultipleUsersPerObjectMixin:
    """
    Mixin to add the deprecated allowMultipleUsersPerObject field to a model.
    May be removed in the future.
    """

    allowMultipleUsersPerObject: Annotated[
        bool,
        Field(
            deprecated=deprecated(
                'Attribute "allowMultipleUsersPerObject" was used in the past to allow multiple users per object but is now deprecated.'
            ),
            exclude=True,
            default=False,
        ),
    ]


class DeprecatedVersionFieldMixin:
    """
    Mixin to add the deprecated version field to a model.
    May be removed in the future.
    """

    version: Annotated[
        str | None,
        Field(
            deprecated=deprecated(
                'Attribute "version" was used in the past to specify the version of the resource but is now deprecated.'
            ),
            exclude=True,
            default=None,
        ),
    ]


class LatLongPoint(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/LatLongPoint

    To avoid circular imports this model is a duplicate of the one in location.py

    Deprecated!
    """

    # inherits kind (deprecated)
    latitude: float = Field(ge=-90.0, le=90.0)
    longitude: float = Field(ge=-180.0, le=180.0)


class DeprecatedLocationsFieldMixin:
    """
    Mixin to add the deprecated locations field to a model.
    May be removed in the future.
    """

    locations: Annotated[
        LatLongPoint | None,
        Field(
            deprecated=deprecated(
                'Attribute "locations" was used in the past to specify locations but is now deprecated.'
            ),
            exclude=True,
            default=None,
        ),
    ]


class TranslatedString(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/LocalizedString#translatedstring

    To avoid circular imports this model is a duplicate of the one in location.py

    Deprecated!
    """

    # inherits kind (deprecated)
    language: str | None = Field(default=None)
    value: str | None = Field(default=None)


class LocalizedString(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/LocalizedString

    To avoid circular imports this model is a duplicate of the one in localized_string.py

    Deprecated!
    """

    # inherits kind (deprecated)
    translatedValues: list[TranslatedString] = Field(default_factory=list)
    defaultValue: TranslatedString


class LabelValue(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/InfoModuleData#labelvalue

    To avoid circular imports this model is a duplicate of the one in data.py

    Deprecated!
    """

    label: str | None = None
    value: str | None = None
    localizedLabel: LocalizedString | None = None
    localizedValue: LocalizedString | None = None


class LabelValueRow(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/InfoModuleData#labelvaluerow

    To avoid circular imports this model is a duplicate of the one in data.py

    Deprecated!
    """

    columns: list[LabelValue] | None = None


class InfoModuleData(Model):
    """
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/InfoModuleData

    To avoid circular imports this model is a duplicate of the one in data.py

    Deprecated!
    """

    labelValueRows: list[LabelValueRow]
    showLastUpdateTime: Annotated[
        bool,
        Field(
            deprecated=deprecated(
                'The Attribute "showLastUpdateTime" on "InfoModuleData" is deprecated.'
            ),
            exclude=True,
            default=False,
        ),
    ]


class DeprecatedInfoModuleDataFieldMixin:
    """
    Mixin to add the deprecated infoModuleData field to a model.
    May be removed in the future.
    """

    infoModuleData: Annotated[
        InfoModuleData | None,
        Field(
            deprecated=deprecated(
                'Attribute "infoModuleData" was used in the past to specify info module data but is now deprecated.'
            ),
            exclude=True,
            default=None,
        ),
    ]


class ImageUri(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Image#imageuri

    To avoid circular imports this model is a duplicate of the one in general.py

    Deprecated!
    """

    uri: AnyUrl
    description: Annotated[
        str | None,
        Field(
            deprecated=deprecated(
                'The Attribute "description" is deprecated on "ImageUri", use "contentDesceription" on parent "Image" Instance.'
            ),
            default=None,
            exclude=True,
        ),
    ]
    localizedDescription: Annotated[
        LocalizedString | None,
        Field(
            deprecated=deprecated(
                'The Attribute "localizedDescription" is deprecated on "ImageUri", use "contentDesceription" on parent "Image" Instance.'
            ),
            default=None,
            exclude=True,
        ),
    ]


class Image(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/Image

    To avoid circular imports this model is a duplicate of the one in general.py

    Deprecated!
    """

    # inherits kind (deprecated)
    sourceUri: ImageUri
    contentDescription: LocalizedString | None = None


class DeprecatedWordMarkFieldMixin:
    """
    Mixin to add the deprecated wordMark field to a model.
    May be removed in the future.
    """

    wordMark: Annotated[
        list[Image] | None,
        Field(
            deprecated=deprecated(
                'Attribute "wordMark" was used in the past to specify the word mark but is now deprecated.'
            ),
            exclude=True,
            default=None,
        ),
    ]



from .session import session_manager
from cryptography.fernet import Fernet


def encrypt_data(data: str) -> str:
    """Encrypt string using the Fernet symmetric encryption algorithm.

    It creates a base64 encoded string.

    The key is fetched from the settings.

    see https://cryptography.io/en/latest/fernet/
    """
    key = session_manager.settings.fernet_encryption_key.encode("utf8")
    fernet = Fernet(key)
    return fernet.encrypt(data.encode("utf8")).decode("utf8")


def decrypt_data(data: str) -> str:
    """Decrypt string using the Fernet symmetric decryption algorithm.

    It takes the base64 encoded string and returns the original value

    The key is fetched from the settings.

    see https://cryptography.io/en/latest/fernet/
    """
    key = session_manager.settings.fernet_encryption_key.encode("utf8")
    fernet = Fernet(key)
    return fernet.decrypt(data.encode("utf8")).decode("utf8")


def generate_fernet_key():
    """Create a new Fernet key."""
    print(Fernet.generate_key().decode("utf8"))



from ..bases import Model
from .general import Image
from .general import Uri
from .localized_string import LocalizedString
from pydantic import Field
from typing import Annotated
from typing_extensions import deprecated


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class TextModuleData(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/TextModuleData

    Google accepts both header and localizedHeader, body and localizedBody
    but if localizedHeader and localizedBody are present, header and body are ignored and set to None.
    """

    header: str | None = Field(max_length=35, default=None)
    body: str | None = Field(max_length=500, default=None)
    localizedHeader: LocalizedString | None = None
    localizedBody: LocalizedString | None = None
    id: str | None = None
    # TODO: custom field validator for id with Field(pattern="^[a-zA-Z0-9_-]+$") plus None
    # TODO: custom modelvalidator for either header/body or localizedHeader/localizedBody


class LinksModuleData(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/LinksModuleData
    """

    uris: list[Uri] | None = None


class ImageModuleData(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/ImageModuleData
    """

    mainImage: Image | None = None
    id: str | None = None


class AppTarget(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/AppLinkData#apptarget
    """

    targetUri: Uri | None = None
    packageName: str | None = None


class AppLinkInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/AppLinkData#applinkinfo
    """

    appLogoImage: Annotated[
        Image | None,
        Field(
            deprecated=deprecated(
                'The Attribute "appLogoImage" on "AppLinkInfo" is deprecated. Image isn\'t supported in the app link module'
            ),
            exclude=True,
            default=None,
        ),
    ]
    title: Annotated[
        LocalizedString | None,
        Field(
            deprecated=deprecated(
                'The Attribute "title" on "AppLinkInfo" is deprecated. Title isn\'t supported in the app link module'
            ),
            exclude=True,
            default=None,
        ),
    ]
    description: Annotated[
        LocalizedString | None,
        Field(
            deprecated=deprecated(
                'The Attribute "description" on "AppLinkInfo" is deprecated. Description isn\'t supported in the app link module'
            ),
            exclude=True,
            default=None,
        ),
    ]
    appTarget: AppTarget | None = None


class AppLinkData(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/AppLinkData
    """

    androidAppLinkInfo: AppLinkInfo | None = None
    iosAppLinkInfo: Annotated[
        AppLinkInfo | None,
        Field(
            deprecated=deprecated(
                'The Attribute "iosAppLinkInfo" on "AppLinkData" is deprecated. Links to open iOS apps are not supported'
            ),
            exclude=True,
            default=None,
        ),
    ]
    webAppLinkInfo: AppLinkInfo | None = None
    displayText: LocalizedString | None = None



from edutap.wallet_google.models.handlers import CallbackData

import json
import pytest


def test_google_public_key_cached_empty(mock_settings):
    from edutap.wallet_google.handlers.validate import google_root_signing_public_keys

    assert google_root_signing_public_keys(mock_settings.google_environment) is not None

    from edutap.wallet_google.handlers.validate import (
        GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE,
    )

    assert (
        GOOGLE_ROOT_SIGNING_PUBLIC_KEYS_VALUE.get(
            mock_settings.google_environment, None
        )
        is not None
    )
    assert google_root_signing_public_keys(mock_settings.google_environment) is not None


callback_data_for_test_failure = {
    "signature": "foo",
    "intermediateSigningKey": {
        "signedKey": {"keyValue": "baz", "keyExpiration": 0},
        "signatures": [
            "MEUCIQD3IATpRM45gpno9Remtx/FiDCOJUp45+C+Qzw6IrgphwIgJijXISc+Ft8Sj9eXNowzuYyXyWlgKAE+tVnN24Sek5M="
        ],
    },
    "protocolVersion": "ECv2SigningOnly",
    "signedMessage": json.dumps(
        {
            "classId": "1",
            "objectId": "2",
            "expTimeMillis": 0,
            "eventType": "save",
            "count": 0,
            "nonce": "fooo",
        }
    ),
}


@pytest.mark.skip(reason="Not implemented")
def test_handler_validate_invalid(mock_settings):
    from edutap.wallet_google.handlers.validate import verified_signed_message

    mock_settings.handler_callback_verify_signature = "1"

    data = CallbackData.model_validate(callback_data_for_test_failure)
    with pytest.raises(Exception):
        verified_signed_message(data)


callback_data = {
    "signature": "MEYCIQCyuBQo/Dao7yUBDUWK12ATFBDkUfJUnropjOaPbPiKEwIhAKXNiVrbNmydpEVIxXRz5z36f8HV2Meq/Td6tqt2+DYO",
    "intermediateSigningKey": {
        "signedKey": '{"keyValue":"MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE3JpSX3AU53vH+IpWBdsbqrL7Ey67QSERsDUztFt8q7t7PzVkh14SeYeokI1zSZiVAWnx4tXD1tCPbrvfFGB8OA\u003d\u003d","keyExpiration":"1735023986000"}',
        "signatures": [
            "MEUCIQD3IBTpRM45gpno9Remtx/FiDCOJUp45+C+Qzw6IrgphwIgJijXISc+Ft8Sj9eXNowzuYyXyWlgKAE+tVnN24Sek5M="
        ],
    },
    "protocolVersion": "ECv2SigningOnly",
    "signedMessage": '{"classId":"3388000000022141777.pass.gift.dev.edutap.eu","objectId":"3388000000022141777.9fd4e525-777c-4e0d-878a-b7993e211997","eventType":"save","expTimeMillis":1734366082269,"count":1,"nonce":"c1359b53-f2bb-4e8f-b392-9a560a21a9a0"}',
}


def test_handler_validate_ok(mock_settings):
    from edutap.wallet_google.handlers.validate import verified_signed_message

    mock_settings.handler_callback_verify_signature = "0"

    data = CallbackData.model_validate(callback_data)
    verified_signed_message(data)



from edutap.wallet_google.settings import ROOT_DIR
from edutap.wallet_google.settings import Settings

import pathlib
import pytest


def test_base_settings():
    settings = Settings()

    assert (
        str(settings.api_url) == "https://walletobjects.googleapis.com/walletobjects/v1"
    )
    assert str(settings.save_url) == "https://pay.google.com/gp/v/save"
    assert [str(s) for s in settings.credentials_scopes] == [
        "https://www.googleapis.com/auth/wallet_object.issuer"
    ]


def test_local_settings(monkeypatch):
    monkeypatch.setenv(
        "EDUTAP_WALLET_GOOGLE_TEST_ISSUER_ID",
        "1234567890123456789",
    )
    monkeypatch.setenv(
        "EDUTAP_WALLET_GOOGLE_CREDENTIALS_FILE",
        str(ROOT_DIR / "tests" / "data" / "credentials_fake.json"),
    )

    settings = Settings()

    assert settings.test_issuer_id == "1234567890123456789"
    assert settings.credentials_file.exists()


def test_settings_cached(mock_settings):
    mock_settings.cached_credentials_info = "test"
    assert mock_settings.credentials_info == "test"


def test_settings_no_credentials_file(mock_settings):
    mock_settings.credentials_file = pathlib.Path("nonexistent.json")
    with pytest.raises(FileNotFoundError):
        mock_settings.credentials_info


def test_settings_wrong_credentials_file(mock_settings):
    mock_settings.credentials_file = (
        ROOT_DIR / "tests" / "data" / "credentials_fake_wrong.json"
    )
    with pytest.raises(ValueError):
        mock_settings.credentials_info



from ..bases import Model


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class ExpiryNotification(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/genericobject#expirynotification
    """

    enableNotification: bool = False


class UpcomingNotification(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/genericobject#upcomingnotification
    """

    enableNotification: bool = False


class Notifications(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/genericobject#notifications
    """

    expiryNotification: ExpiryNotification | None = None
    upcomingNotification: UpcomingNotification | None = None



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from .datetime import TimeInterval
from .enums import MessageType
from .localized_string import LocalizedString


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Message(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/Message
    """

    # inherits kind (deprecated)
    header: str | None = None
    body: str | None = None
    displayInterval: TimeInterval | None = None
    id: str | None = None
    messageType: MessageType = MessageType.MESSAGE_TYPE_UNSPECIFIED
    localizedHeader: LocalizedString | None = None
    localizedBody: LocalizedString | None = None



from fastapi import FastAPI
from fastapi.testclient import TestClient


# this callback data can be verified given the credentials.json from demo.edutap.eu is provided.
real_callback_data = {
    "signature": "MEYCIQCyuBQo/Dao7yUBDUWK12ATFBDkUfJUnropjOaPbPiKEwIhAKXNiVrbNmydpEVIxXRz5z36f8HV2Meq/Td6tqt2+DYO",
    "intermediateSigningKey": {
        "signedKey": '{"keyValue":"MFkwEwYHKoZIzj0CAQYIKoZIzj0DAQcDQgAE3JpSX3AU53vH+IpWBdsbqrL7Ey67QSERsDUztFt8q7t7PzVkh14SeYeokI1zSZiVAWnx4tXD1tCPbrvfFGB8OA\\u003d\\u003d","keyExpiration":"1735023986000"}',
        "signatures": [
            "MEUCIQD3IBTpRM45gpno9Remtx/FiDCOJUp45+C+Qzw6IrgphwIgJijXISc+Ft8Sj9eXNowzuYyXyWlgKAE+tVnN24Sek5M="
        ],
    },
    "protocolVersion": "ECv2SigningOnly",
    "signedMessage": '{"classId":"3388000000022141777.pass.gift.dev.edutap.eu","objectId":"3388000000022141777.9fd4e525-777c-4e0d-878a-b7993e211997","eventType":"save","expTimeMillis":1734366082269,"count":1,"nonce":"c1359b53-f2bb-4e8f-b392-9a560a21a9a0"}',
}


def test_callback_disabled_signature_check_OK(mock_settings):
    from edutap.wallet_google.models.handlers import CallbackData

    # test callback handler without signature check
    mock_settings.handler_callback_verify_signature = "0"

    callback_data = CallbackData(**real_callback_data)

    from edutap.wallet_google.handlers.fastapi import router

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.post(
        "/wallet/google/callback", json=callback_data.model_dump(mode="json")
    )
    assert resp.status_code == 200
    assert resp.json() == {"status": "success"}


def test_callback_disabled_signature_check_ERROR(mock_settings):
    from edutap.wallet_google.models.handlers import CallbackData

    # test callback handler without signature check
    mock_settings.handler_callback_verify_signature = "0"

    callback_data = CallbackData(**real_callback_data)
    callback_data.signedMessage = '{"classId":"1.x","objectId":"1.y","eventType":"save","expTimeMillis":1734366082269,"count":1,"nonce":""}'

    from edutap.wallet_google.handlers.fastapi import router

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.post(
        "/wallet/google/callback", json=callback_data.model_dump(mode="json")
    )
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Error while handling the callbacks (exception)."}'


def test_callback_disabled_signature_check_NOTIMPLEMENTED(monkeypatch, mock_settings):
    from edutap.wallet_google.models.handlers import CallbackData

    # test callback handler without signature check
    mock_settings.handler_callback_verify_signature = "0"

    def raise_not_implemented():
        raise NotImplementedError

    # disable all plugins for callback_handlers - patch the instance at fastapi!
    monkeypatch.setattr(
        "edutap.wallet_google.handlers.fastapi.get_callback_handlers",
        raise_not_implemented,
    )

    callback_data = CallbackData(**real_callback_data)
    callback_data.signedMessage = '{"classId":"1.x","objectId":"1.y","eventType":"save","expTimeMillis":1734366082269,"count":1,"nonce":"abcde"}'

    from edutap.wallet_google.handlers.fastapi import router

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.post(
        "/wallet/google/callback", json=callback_data.model_dump(mode="json")
    )
    assert resp.status_code == 500
    assert resp.text == '{"detail":"No callback handlers were registered."}'


def test_callback_disabled_signature_check_TIMEOUT(mock_settings):
    from edutap.wallet_google.models.handlers import CallbackData

    # test callback handler without signature check
    mock_settings.handler_callback_verify_signature = "0"
    # set low timeout to trigger a timeout cancellation
    mock_settings.handlers_callback_timeout = 0.1

    callback_data = CallbackData(**real_callback_data)
    callback_data.signedMessage = '{"classId":"TIMEOUT.x","objectId":"1.x","eventType":"save","expTimeMillis":250,"count":1,"nonce":"abcde"}'

    from edutap.wallet_google.handlers.fastapi import router

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.post(
        "/wallet/google/callback", json=callback_data.model_dump(mode="json")
    )
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Error while handling the callbacks (timeout)."}'


def test_image_OK(mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('OK')}")
    assert resp.status_code == 200
    assert resp.text == "mock-a-jepg"


def test_image_ERROR(mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('ERROR')}")
    assert resp.status_code == 404
    assert resp.text == '{"detail":"Image not found."}'


def test_image_TIMEOUT(mock_settings, mock_fernet_encryption_key):
    mock_settings.handlers_image_timeout = 0.1

    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('TIMEOUT')}")
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Error while handling the image (timeout)."}'


def test_image_CANCEL(mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('CANCEL')}")
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Error while handling the image (cancel)."}'


def test_image_UNEXPECTED(mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('UNEXPECTED')}")
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Error while handling the image (exception)."}'


def test_image_NOTIMPLEMENTED(monkeypatch, mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    def raise_not_implemented():
        raise NotImplementedError

    # disable all plugins for callback_handlers - patch the instance at fastapi!
    monkeypatch.setattr(
        "edutap.wallet_google.handlers.fastapi.get_image_providers",
        raise_not_implemented,
    )

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('ANYWAY')}")
    assert resp.status_code == 500
    assert resp.text == '{"detail":"No image providers were registered."}'


def test_image_TO_MANY_REGISTERED(monkeypatch, mock_fernet_encryption_key):
    from edutap.wallet_google.handlers.fastapi import router
    from edutap.wallet_google.utils import encrypt_data

    # disable all plugins for callback_handlers - patch the instance at fastapi!
    monkeypatch.setattr(
        "edutap.wallet_google.handlers.fastapi.get_image_providers", lambda: [1, 2]
    )

    app = FastAPI()
    app.include_router(router)
    client = TestClient(app)
    resp = client.get(f"/wallet/google/images/{encrypt_data('ANYWAY')}")
    assert resp.status_code == 500
    assert resp.text == '{"detail":"Multiple image providers found, abort."}'



import datetime
import pytest


def test_create_payload():

    from edutap.wallet_google import api

    payload = api._create_payload(
        [
            api.new(
                "Reference", {"id": "test-1.edutap.eu", "model_name": "GenericObject"}
            ),
            api.new(
                "OfferObject",
                {"id": "test-2.edutap.eu", "classId": "test-class-1.edutap.eu"},
            ),
        ]
    )
    expected = '{"offerObjects":[{"id":"test-2.edutap.eu","classId":"test-class-1.edutap.eu","state":"STATE_UNSPECIFIED","hasLinkedDevice":false,"disableExpirationNotification":false,"notifyPreference":"NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED"}],"genericObjects":[{"id":"test-1.edutap.eu"}]}'
    assert payload.model_dump_json(exclude_none=True) == expected


def test_create_claims():

    from edutap.wallet_google import api

    models = [
        api.new("Reference", {"id": "test-1.edutap.eu", "model_name": "GenericObject"}),
        api.new(
            "OfferObject",
            {"id": "test-2.edutap.eu", "classId": "test-class-1.edutap.eu"},
        ),
    ]
    expected = {
        "iss": "test@example.com",
        "aud": "google",
        "typ": "savetowallet",
        "iat": "1737115974",
        "exp": "",
        "payload": {
            "offerObjects": [
                {
                    "id": "test-2.edutap.eu",
                    "classId": "test-class-1.edutap.eu",
                    "state": "STATE_UNSPECIFIED",
                    "hasLinkedDevice": False,
                    "disableExpirationNotification": False,
                    "notifyPreference": "NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED",
                }
            ],
            "genericObjects": [{"id": "test-1.edutap.eu"}],
        },
        "origins": [],
    }

    claims = api._create_claims(
        "test@example.com",
        [],
        models,
        iat=datetime.datetime(2025, 1, 17, 12, 12, 54, 0, datetime.timezone.utc),
        exp="",
    )

    dumped = claims.model_dump(
        mode="json",
        exclude_unset=False,
        exclude_defaults=False,
        exclude_none=True,
    )
    assert expected == dumped


def test_api_save_link(mock_settings):
    from edutap.wallet_google.settings import ROOT_DIR

    mock_settings.credentials_file = (
        ROOT_DIR / "tests" / "data" / "credentials_fake.json"
    )

    from edutap.wallet_google import api

    link = api.save_link(
        [
            api.new(
                "Reference",
                {
                    "id": "1234567890123456789.test-1.edutap.eu",
                    "model_name": "GenericObject",
                },
            ),
            api.new(
                "OfferObject",
                {
                    "id": "1234567890123456789.test-2.edutap.eu",
                    "classId": "1234567890123456789.test-class-1.edutap.eu",
                },
            ),
        ],
        iat=datetime.datetime(2025, 1, 22, 10, 20, 0, 0, datetime.timezone.utc),
    )
    expected = "https://pay.google.com/gp/v/save/eyJ0eXAiOiAiSldUIiwgImFsZyI6ICJSUzI1NiIsICJraWQiOiAiMTIzNDU2Nzg5MGFiY2RlZjEyMzQ1Njc4OTBhYmNkZWYxMjM0NTY3OCJ9.eyJpc3MiOiAiZWR1dGFwLXRlc3QtZXhhbXBsZUBzb2RpdW0tcmF5LTEyMzQ1Ni5pYW0uZ3NlcnZpY2VhY2NvdW50LmNvbSIsICJhdWQiOiAiZ29vZ2xlIiwgInR5cCI6ICJzYXZldG93YWxsZXQiLCAiaWF0IjogIjE3Mzc1NDEyMDAiLCAicGF5bG9hZCI6IHsib2ZmZXJPYmplY3RzIjogW3siaWQiOiAiMTIzNDU2Nzg5MDEyMzQ1Njc4OS50ZXN0LTIuZWR1dGFwLmV1IiwgImNsYXNzSWQiOiAiMTIzNDU2Nzg5MDEyMzQ1Njc4OS50ZXN0LWNsYXNzLTEuZWR1dGFwLmV1IiwgInN0YXRlIjogIlNUQVRFX1VOU1BFQ0lGSUVEIiwgImhhc0xpbmtlZERldmljZSI6IGZhbHNlLCAiZGlzYWJsZUV4cGlyYXRpb25Ob3RpZmljYXRpb24iOiBmYWxzZSwgIm5vdGlmeVByZWZlcmVuY2UiOiAiTk9USUZJQ0FUSU9OX1NFVFRJTkdTX0ZPUl9VUERBVEVTX1VOU1BFQ0lGSUVEIn1dLCAiZ2VuZXJpY09iamVjdHMiOiBbeyJpZCI6ICIxMjM0NTY3ODkwMTIzNDU2Nzg5LnRlc3QtMS5lZHV0YXAuZXUifV19LCAib3JpZ2lucyI6IFtdLCAiZXhwIjogIiJ9.M5rHck9hIZOCHyIb9waDHsVHIzIQD-Hle4EK-BJYaRnGWeAnG2Gq9jxG9anBWlujRoqwFBsKozaHagGB9AyzpVDbfPJPY7Zm2jRwhe76zzmkSruEr25H1hf_OMxZgpUoyatASikANPPBrayA9-D9nuNThbNCOuSmwFXv4iDdxHHfCzq0iEKiPvw_6fePF7bdPnkv1uE0GA5rseM2olJ830U7xB7wuw7SA84OPoRclMiCvl3RjisTE2PfrkZ0sJlbxM16-aMkAQ-FxlERLoxkeWgSA2yYXuqjEdH0TVbg_b4tGKgQkTXo2TxhaVam4ZqdjYzTkgu9oesPVVae8ZgEtQ"
    assert link == expected


def test__convert_str_or_datetime_to_str__timestamp():
    from edutap.wallet_google.api import _convert_str_or_datetime_to_str

    dt = datetime.datetime(2021, 1, 1, 12, 0, 0, 0, datetime.timezone.utc)
    expected = "1609502400"

    assert _convert_str_or_datetime_to_str(dt) == expected


def test__convert_str_or_datetime_to_str__str_int_lt_zero():
    from edutap.wallet_google.api import _convert_str_or_datetime_to_str

    with pytest.raises(ValueError):
        _convert_str_or_datetime_to_str("-1")


def test__convert_str_or_datetime_to_str__str_int_tr_4bytes():
    from edutap.wallet_google.api import _convert_str_or_datetime_to_str

    with pytest.raises(ValueError):
        _convert_str_or_datetime_to_str(f"{2**32 + 1}")


def test__convert_str_or_datetime_to_str__not_decimal():
    from edutap.wallet_google.api import _convert_str_or_datetime_to_str

    with pytest.raises(ValueError):
        _convert_str_or_datetime_to_str("x 100")



from ...registry import register_model
from ..bases import WithIdModel
from ..datatypes.barcode import Barcode
from ..datatypes.barcode import RotatingBarcode
from ..datatypes.class_template_info import ClassTemplateInfo
from ..datatypes.data import AppLinkData
from ..datatypes.data import ImageModuleData
from ..datatypes.data import LinksModuleData
from ..datatypes.data import TextModuleData
from ..datatypes.datetime import TimeInterval
from ..datatypes.enums import MultipleDevicesAndHoldersAllowedStatus
from ..datatypes.enums import State
from ..datatypes.enums import ViewUnlockRequirement
from ..datatypes.general import CallbackOptions
from ..datatypes.general import GroupingInfo
from ..datatypes.general import Image
from ..datatypes.general import PassConstraints
from ..datatypes.general import SaveRestrictions
from ..datatypes.general import SecurityAnimation
from ..datatypes.message import Message
from ..datatypes.moduledata import ValueAddedModuleData
from pydantic import Field
from pydantic import model_validator


@register_model(
    "Reference",
    url_part="",
    can_create=False,
    can_read=False,
    can_update=False,
    can_list=False,
    can_message=False,
)
class Reference(WithIdModel):
    """
    References an existing wallet object.

    It is used to create the JWT for the add to wallet link.
    The id must be an existing wallet object id.

    Either model_name or mode_type must be set.
    """

    # inherits id

    # mode_name and model_type are implementation specific for this package
    model_name: str | None = Field(exclude=True, default=None)
    model_type: type[WithIdModel] | None = Field(exclude=True, default=None)

    @model_validator(mode="after")
    def check_one_of(self) -> "Reference":
        if self.model_name is None and self.model_type is None:
            raise ValueError("One of [model_name, model_type] must be set")
        if self.model_name is not None and self.model_type is not None:
            raise ValueError("Only one of [model_name, model_type] must be set")
        return self


class ClassModel(WithIdModel):
    """
    BaseModel for all Google Wallet Class Models.

    Even if not documented explicitly by Google, the GenericClass acts as
    the base-class for all Google Wallet Classes. So this here is the GenericClass.
    To keep the inheritance chain semantically clean, there is a registered GenericClass
    inheriting from here without any added attributes in the module `generic.py`.
    see https://developers.google.com/wallet/generic/rest/v1/genericclass
    """

    # Attribute order as in Google's documentation to make future updates easier!
    # last check: 2024-12-02

    # inherits id
    classTemplateInfo: ClassTemplateInfo | None = None
    imageModulesData: list[ImageModuleData] | None = None
    textModulesData: list[TextModuleData] | None = None
    linksModuleData: LinksModuleData | None = None
    enableSmartTap: bool | None = None
    redemptionIssuers: list[str] | None = None  # string (int64 format)
    securityAnimation: SecurityAnimation | None = None
    multipleDevicesAndHoldersAllowedStatus: MultipleDevicesAndHoldersAllowedStatus = (
        MultipleDevicesAndHoldersAllowedStatus.STATUS_UNSPECIFIED
    )
    callbackOptions: CallbackOptions | None = None
    viewUnlockRequirement: ViewUnlockRequirement = (
        ViewUnlockRequirement.VIEW_UNLOCK_REQUIREMENT_UNSPECIFIED
    )
    messages: list[Message] | None = None
    appLinkData: AppLinkData | None = None
    valueAddedModuleData: ValueAddedModuleData | None = None


class ObjectModel(WithIdModel):
    """
    Base model for all Google Wallet Object models.

    Opposed to the ClassModel, this class does not represent the GenericObject.
    So, it does not act as a base model for all other wallet object types.
    It is just a base model with all attributes all Google Wallet Objects models have in common.
    """

    # inherits id
    classId: str

    # Templating and Visual Data
    imageModulesData: list[ImageModuleData] | None = None
    textModulesData: list[TextModuleData] | None = None
    linksModuleData: LinksModuleData | None = None

    # Security Options
    passConstraints: PassConstraints | None = None

    # Smart Tap Option
    smartTapRedemptionValue: str | None = None

    # Barcode Options
    barcode: Barcode | None = None
    rotatingBarcode: RotatingBarcode | None = None

    # other general
    groupingInfo: GroupingInfo | None = None
    saveRestrictions: SaveRestrictions | None = None
    linkedObjectIds: list[str] | None = None
    valueAddedModuleData: ValueAddedModuleData | None = None
    messages: list[Message] | None = None
    appLinkData: AppLinkData | None = None
    state: State = Field(default=State.STATE_UNSPECIFIED)
    hasUsers: bool | None = None
    validTimeInterval: TimeInterval | None = None


class StyleableMixin:
    """
    Mixin for Google Wallet Classes/Objects that can be styled.
    """

    hexBackgroundColor: str | None = None


class CommonLogosMixin:
    """
    Mixin for Google Wallet Classes/Objects with a common logo.

    Do not use/alias this mixin for Google Wallet Classes, as they have a different logo attribute.
    """

    logo: Image | None = None
    wideLogo: Image | None = None
    heroImage: Image | None = None



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from .enums import DoorsOpenLabel
from .general import LocalizedString

import datetime


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class EventVenue(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#eventvenue
    """

    # inherits kind (deprecated)
    name: LocalizedString | None = None
    address: LocalizedString | None = None


class EventDateTime(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass#eventdatetime
    """

    # inherits kind (deprecated)
    # TODO: can't be properly resolved because we have a custom module named datetime, i.e. global datetime is shadowed
    doorsOpen: datetime.datetime | None = None
    start: datetime.datetime | None = None
    end: datetime.datetime | None = None
    doorsOpenLabel: DoorsOpenLabel | None = None
    customDoorsOpenLabel: LocalizedString | None = None


class EventSeat(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/eventticketobject#eventseat
    """

    # inherits kind (deprecated)
    seat: LocalizedString | None = None
    row: LocalizedString | None = None
    section: LocalizedString | None = None
    gate: LocalizedString | None = None


class EventReservationInfo(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/eventticketobject#eventreservationinfo
    """

    # inherits kind (deprecated)
    confirmationCode: str | None = None



from pathlib import Path
from pydantic import AnyHttpUrl
from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict
from typing import Literal

import json


ENV_PREFIX = "EDUTAP_WALLET_GOOGLE_"
ROOT_DIR = Path(__file__).parent.parent.parent.parent.resolve()
API_URL = "https://walletobjects.googleapis.com/walletobjects/v1"
SAVE_URL = "https://pay.google.com/gp/v/save"
SCOPES = ["https://www.googleapis.com/auth/wallet_object.issuer"]


class Settings(BaseSettings):
    """Settings for Google Wallet Preferences.

    For more on how these settings work follow https://docs.pydantic.dev/latest/concepts/pydantic_settings/

    Any default can be overridden by setting the corresponding environment variable prefixed with `EDUTAP_WALLET_GOOGLE_`.
    If a `.env` file is present in the root directory of the project, the environment variables will be loaded from there.
    """

    model_config = SettingsConfigDict(
        env_prefix=ENV_PREFIX,
        env_file=".env",
        env_file_encoding="utf-8",
        case_sensitive=False,
        extra="ignore",
    )

    record_api_calls_dir: Path | None = None
    api_url: AnyHttpUrl = AnyHttpUrl(API_URL)
    save_url: AnyHttpUrl = AnyHttpUrl(SAVE_URL)

    handler_prefix: str = "/wallet/google"
    handler_prefix_callback: str = ""
    handler_prefix_images: str = ""
    handler_callback_verify_signature: str = "1"
    handler_image_cache_control: str = "no-cache"
    handlers_callback_timeout: float = 5.0
    handlers_image_timeout: float = 5.0

    credentials_file: Path = ROOT_DIR / "tests" / "data" / "credentials_fake.json"
    credentials_scopes: list[str] = SCOPES
    test_issuer_id: str = Field(default="")

    fernet_encryption_key: str = ""

    google_environment: Literal["production", "testing"] = "testing"

    cached_credentials_info: dict[str, str] = {}

    @property
    def credentials_info(self) -> dict[str, str]:
        if credentials_info := self.cached_credentials_info:
            return credentials_info
        if not self.credentials_file.exists():
            raise FileNotFoundError(
                f"EDUTAP_WALLET_GOOGLE_CREDENTIALS_FILE={self.credentials_file} does not exist."
            )
        with open(self.credentials_file) as fp:
            self.cached_credentials_info = json.load(fp)
        if not isinstance(self.cached_credentials_info, dict):
            raise ValueError(
                f"EDUTAP_WALLET_GOOGLE_CREDENTIALS_FILE={self.credentials_file} content is not a dict"
            )
        return self.cached_credentials_info
